# This file is a part of the AnyBlok project
#
#    Copyright (C) 2015 Georges Racinet <gracinet@anybox.fr>
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file,You can
# obtain one at http://mozilla.org/MPL/2.0/.
"""Authorization subframework.

The founding principle of authorization handling within Anyblok is to check
authorization explicitely at the edge of the system (for instance,
for applications
exposed over HTTP, that would be in the controllers), because that is where the
idea of user, or slightly more generally, has functional semantics that can
be interpreted in the context of a given action.

In that spirit, we don't pass the user to the core framework and business
layers.
Instead, these provide *policies* to check permissions on records or query
records according to the.

The declarations at the edge will *associate* the policies with the
models. The edge user-aware methods will call the check and query facilities
provided by the core that themselves apply the relevant policies.
"""



# This file is a part of the AnyBlok project
#
#    Copyright (C) 2015 Jean-Sebastien SUZANNE <jssuzanne@anybox.fr>
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file,You can
# obtain one at http://mozilla.org/MPL/2.0/.
from anyblok.tests.testcase import BlokTestCase


class TestQuery(BlokTestCase):

    def test_dictone(self):
        query = self.registry.System.Model.query().filter_by(
            name='Model.System.Blok')
        model = query.one()
        self.assertEqual(query.dictone(), {
            'name': model.name,
            'table': model.table,
            'is_sql_model': model.is_sql_model,
            'description': model.description,
        })

    def test_dictone_on_some_column(self):
        query = self.registry.System.Model.query('name', 'table').filter_by(
            name='Model.System.Blok')
        model = query.one()
        self.assertEqual(query.dictone(), {
            'name': model.name,
            'table': model.table,
        })

    def test_dictone_on_some_column_with_label(self):
        M = self.registry.System.Model
        query = M.query(M.name.label('n2'), M.table.label('t2')).filter_by(
            name='Model.System.Blok')
        model = query.one()
        self.assertEqual(query.dictone(), {
            'n2': model.n2,
            't2': model.t2,
        })

    def test_dictfirst(self):
        query = self.registry.System.Model.query()
        model = query.first()
        self.assertEqual(query.dictfirst(), {
            'name': model.name,
            'table': model.table,
            'is_sql_model': model.is_sql_model,
            'description': model.description,
        })

    def test_dictfirst_on_some_column(self):
        query = self.registry.System.Model.query('name', 'table')
        model = query.first()
        self.assertEqual(query.dictfirst(), {
            'name': model.name,
            'table': model.table,
        })

    def test_dictfirst_on_some_column_with_label(self):
        M = self.registry.System.Model
        query = M.query(M.name.label('n2'), M.table.label('t2'))
        model = query.first()
        self.assertEqual(query.dictfirst(), {
            'n2': model.n2,
            't2': model.t2,
        })

    def test_dictall(self):
        query = self.registry.System.Model.query().limit(2)
        models = query.all()

        def to_dict(model):
            return {
                'name': model.name,
                'table': model.table,
                'is_sql_model': model.is_sql_model,
                'description': model.description,
            }

        dictall = query.dictall()
        for i in range(2):
            self.assertIn(to_dict(models[i]), dictall)

    def test_dictall_on_some_column(self):
        query = self.registry.System.Model.query('name', 'table').limit(2)
        models = query.all()

        def to_dict(model):
            return {
                'name': model.name,
                'table': model.table,
            }

        dictall = query.dictall()
        for i in range(2):
            self.assertIn(to_dict(models[i]), dictall)

    def test_dictall_on_some_column_with_label(self):
        M = self.registry.System.Model
        query = M.query(M.name.label('n2'), M.table.label('t2')).limit(2)
        models = query.all()

        def to_dict(model):
            return {
                'n2': model.n2,
                't2': model.t2,
            }

        dictall = query.dictall()
        for i in range(2):
            self.assertIn(to_dict(models[i]), dictall)



# This file is a part of the AnyBlok project
#
#    Copyright (C) 2014 Jean-Sebastien SUZANNE <jssuzanne@anybox.fr>
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file,You can
# obtain one at http://mozilla.org/MPL/2.0/.
from anyblok.tests.testcase import DBTestCase
from anyblok.declarations import Declarations, listen
from anyblok.column import Integer, Boolean, String


register = Declarations.register
Model = Declarations.Model
Mixin = Declarations.Mixin
Core = Declarations.Core


class TestEvent(DBTestCase):

    def check_event(self, registry):
        self.assertEqual(registry.Test.x, 0)
        registry.Event.fire('fireevent')
        self.assertEqual(registry.Test.x, 1)
        registry.Event.fire('fireevent', a=2, b=2)
        self.assertEqual(registry.Test.x, 4)
        registry.Event.fire('fireevent')
        self.assertEqual(registry.Test.x, 1)

    def test_simple_event_from_model(self):

        def add_in_registry():

            @register(Model)
            class Event:
                pass

            @register(Model)
            class Test:

                x = 0

                @listen(Model.Event, 'fireevent')
                def my_event(cls, a=1, b=1):
                    cls.x = a * b

        registry = self.init_registry(add_in_registry)
        self.check_event(registry)

    def test_simple_event_from_model_by_name(self):

        def add_in_registry():

            @register(Model)
            class Event:
                pass

            @register(Model)
            class Test:

                x = 0

                @listen('Model.Event', 'fireevent')
                def my_event(cls, a=1, b=1):
                    cls.x = a * b

        registry = self.init_registry(add_in_registry)
        self.assertEqual(len(registry.events['Model.Event']['fireevent']),
                         1)
        self.check_event(registry)

    def test_simple_event_from_mixin(self):

        def add_in_registry():

            @register(Model)
            class Event:
                pass

            @register(Mixin)
            class MTest:

                x = 0

                @listen(Model.Event, 'fireevent')
                def my_event(cls, a=1, b=1):
                    cls.x = a * b

            @register(Model)
            class Test(Mixin.MTest):
                pass

        registry = self.init_registry(add_in_registry)
        self.assertEqual(len(registry.events['Model.Event']['fireevent']),
                         1)
        self.check_event(registry)

    def test_simple_event_from_core(self):

        def add_in_registry():

            @register(Model)
            class Event:
                pass

            @register(Core)
            class Base:

                x = 0

                @listen(Model.Event, 'fireevent')
                def my_event(cls, a=1, b=1):
                    cls.x = a * b

            @register(Model)
            class Test:
                pass

        registry = self.init_registry(add_in_registry)
        self.check_event(registry)

    def add_in_registry_inherited(self, withcore=False, withmixin=False,
                                  withmodel=False):

        @register(Model)
        class Event:
            pass

        @register(Core)
        class Base:

            x = 0

            if withcore:
                @listen(Model.Event, 'fireevent')
                def my_event(cls, a=1, b=1):
                    pass

        @register(Mixin)
        class MTest:

            if withmixin:
                @listen(Model.Event, 'fireevent')
                def my_event(cls, a=1, b=1):
                    pass

        @register(Model)
        class Test(Mixin.MTest):

            if withmodel:
                @listen(Model.Event, 'fireevent')
                def my_event(cls, a=1, b=1):
                    cls.x = a * b
            else:
                @classmethod
                def my_event(cls, a=1, b=1):
                    cls.x = a * b

    def test_inherited_without_event(self):
        registry = self.init_registry(self.add_in_registry_inherited)
        self.assertEqual(registry.Test.x, 0)
        registry.Event.fire('fireevent')
        self.assertEqual(registry.Test.x, 0)
        registry.Event.fire('fireevent', a=2, b=2)
        self.assertEqual(registry.Test.x, 0)
        registry.Event.fire('fireevent')
        self.assertEqual(registry.Test.x, 0)

    def test_inherited_with_event_on_core(self):
        registry = self.init_registry(self.add_in_registry_inherited,
                                      withcore=True)
        self.check_event(registry)

    def test_inherited_with_event_on_mixin(self):
        registry = self.init_registry(self.add_in_registry_inherited,
                                      withmixin=True)
        self.assertEqual(
            len(registry.events['Model.Event']['fireevent']), 1)
        self.check_event(registry)

    def test_inherited_with_event_on_core_and_mixin(self):
        registry = self.init_registry(self.add_in_registry_inherited,
                                      withcore=True, withmixin=True)
        self.check_event(registry)

    def test_inherited_with_event_on_mixin_and_model(self):
        registry = self.init_registry(self.add_in_registry_inherited,
                                      withmodel=True, withmixin=True)
        self.assertEqual(len(registry.events['Model.Event']['fireevent']), 1)
        self.check_event(registry)

    def test_sqlalchemy_listen_on_model(self):

        def add_in_registry():

            @register(Model)
            class Test:

                id = Integer(primary_key=True)
                val = Boolean(default=False)

                @listen('Model.Test', 'before_insert')
                def my_event(cls, mapper, connection, target):
                    target.val = True

        registry = self.init_registry(add_in_registry)
        t = registry.Test.insert()
        self.assertTrue(t.val)

    def test_sqlalchemy_listen_on_column(self):

        def add_in_registry():

            @register(Model)
            class Test:

                id = Integer(primary_key=True)
                val = String()

                @listen('Model.Test=>val', 'set', retval=True)
                def my_event(cls, target, value, oldvalue, initiator):
                    return 'test_' + value

        registry = self.init_registry(add_in_registry)
        t = registry.Test.insert()
        self.assertIsNone(t.val)
        t.val = 'test'
        registry.flush()
        self.assertEqual(t.val, 'test_test')



# This file is a part of the AnyBlok project
#
#    Copyright (C) 2015 Jean-Sebastien SUZANNE <jssuzanne@anybox.fr>
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file,You can
# obtain one at http://mozilla.org/MPL/2.0/.
from anyblok.tests.testcase import BlokTestCase
from ..exceptions import ParameterException


class TestSystemParameter(BlokTestCase):

    def test_set(self):
        Parameter = self.registry.System.Parameter
        query = Parameter.query().filter(Parameter.key == 'test.parameter')
        if query.count():
            self.fail('key for test already existing')

        Parameter.set('test.parameter', True)
        self.assertEqual(query.count(), 1)
        self.assertEqual(query.first().value, {'value': True})
        self.assertEqual(query.first().multi, False)

    def test_set_with_multi(self):
        Parameter = self.registry.System.Parameter
        query = Parameter.query().filter(Parameter.key == 'test.parameter')
        if query.count():
            self.fail('key for test already existing')

        Parameter.set('test.parameter', {'test': True})
        self.assertEqual(query.count(), 1)
        self.assertEqual(query.first().value, {'test': True})
        self.assertEqual(query.first().multi, True)

    def test_get(self):
        Parameter = self.registry.System.Parameter
        Parameter.set('test.parameter', True)
        self.assertEqual(Parameter.get('test.parameter'), True)

    def test_get_with_multi(self):
        Parameter = self.registry.System.Parameter
        Parameter.set('test.parameter', {'test': True})
        self.assertEqual(Parameter.get('test.parameter'), {'test': True})

    def test_unexisting_get(self):
        Parameter = self.registry.System.Parameter
        with self.assertRaises(ParameterException):
            Parameter.get('test.parameter')

    def test_count(self):
        Parameter = self.registry.System.Parameter
        self.assertEqual(Parameter.is_exist('test.parameter'), False)
        Parameter.set('test.parameter', True)
        self.assertEqual(Parameter.is_exist('test.parameter'), True)

    def test_set_existing_key(self):
        Parameter = self.registry.System.Parameter
        query = Parameter.query().filter(Parameter.key == 'test.parameter')
        self.assertEqual(query.count(), 0)
        Parameter.set('test.parameter', True)
        self.assertEqual(query.count(), 1)
        self.assertEqual(Parameter.get('test.parameter'), True)
        Parameter.set('test.parameter', False)
        self.assertEqual(query.count(), 1)
        self.assertEqual(Parameter.get('test.parameter'), False)



# This file is a part of the AnyBlok project
#
#    Copyright (C) 2014 Jean-Sebastien SUZANNE <jssuzanne@anybox.fr>
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file,You can
# obtain one at http://mozilla.org/MPL/2.0/.
from anyblok.declarations import Declarations, listen
from anyblok.field import Function
from anyblok.column import String, Boolean
from logging import getLogger

logger = getLogger(__name__)

register = Declarations.register
System = Declarations.Model.System


@register(System)
class Model:
    """Models assembled"""

    def __str__(self):
        if self.description:
            return self.description

        return self.name

    name = String(size=256, primary_key=True)
    table = String(size=256)
    is_sql_model = Boolean(label="Is a SQL model")
    description = Function(fget='get_model_doc_string')

    def get_model_doc_string(self):
        """ Return the docstring of the model
        """
        m = self.registry.get(self.name)
        if hasattr(m, '__doc__'):
            return m.__doc__ or ''

        return ''

    @listen('Model.System.Model', 'Update Model')
    def listener_update_model(cls, model):
        cls.registry.System.Cache.invalidate(model, '_fields_description')
        cls.registry.System.Cache.invalidate(
            model, 'find_remote_attribute_to_expire')
        cls.registry.System.Cache.invalidate(
            model, 'get_hybrid_property_columns')

    @classmethod
    def get_field_model(cls, field):
        ftype = field.property.__class__.__name__
        if ftype == 'ColumnProperty':
            return cls.registry.System.Column
        elif ftype == 'RelationshipProperty':
            return cls.registry.System.RelationShip
        else:
            raise Exception('Not implemented yet')

    @classmethod
    def get_field(cls, model, cname):
        if cname in model.loaded_fields.keys():
            field = model.loaded_fields[cname]
            Field = cls.registry.System.Field
        else:
            field = getattr(model, cname)
            Field = cls.get_field_model(field)

        return field, Field

    @classmethod
    def update_fields(cls, model, table):
        fsp = cls.registry.loaded_namespaces_first_step
        m = cls.registry.get(model)
        for cname in m.loaded_columns:
            ftype = fsp[model][cname].__class__.__name__
            field, Field = cls.get_field(m, cname)
            cname = Field.get_cname(field, cname)
            query = Field.query()
            query = query.filter(Field.model == model)
            query = query.filter(Field.name == cname)
            if query.count():
                Field.alter_field(query.first(), field, ftype)
            else:
                Field.add_field(cname, field, model, table, ftype)

    @classmethod
    def add_fields(cls, model, table):
        fsp = cls.registry.loaded_namespaces_first_step
        m = cls.registry.get(model)
        is_sql_model = len(m.loaded_columns) > 0
        cls.insert(name=model, table=table,
                   is_sql_model=is_sql_model)
        for cname in m.loaded_columns:
            field, Field = cls.get_field(m, cname)
            cname = Field.get_cname(field, cname)
            ftype = fsp[model][cname].__class__.__name__
            Field.add_field(cname, field, model, table, ftype)

    @classmethod
    def update_list(cls):
        """ Insert and update the table of models

        :exception: Exception
        """
        for model in cls.registry.loaded_namespaces.keys():
            try:
                # TODO need refactor, then try except pass whenever refactor
                # not apply
                m = cls.registry.get(model)
                table = ''
                if hasattr(m, '__tablename__'):
                    table = m.__tablename__

                _m = cls.query('name').filter(cls.name == model)
                if _m.count():
                    cls.update_fields(model, table)
                else:
                    cls.add_fields(model, table)

                if m.loaded_columns:
                    cls.fire('Update Model', model)

            except Exception as e:
                logger.exception(str(e))



# This file is a part of the AnyBlok project
#
#    Copyright (C) 2014 Jean-Sebastien SUZANNE <jssuzanne@anybox.fr>
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file,You can
# obtain one at http://mozilla.org/MPL/2.0/.
from anyblok.tests.testcase import TestCase, DBTestCase
from sqlalchemy import Integer as SA_Integer
from sqlalchemy.exc import StatementError
from anyblok import Declarations
from anyblok.field import FieldException
from anyblok.column import (Column, Boolean, Json, String, BigInteger,
                            SmallInteger, uString, Text, uText, Selection,
                            Date, DateTime, Time, Interval, Decimal, Float,
                            LargeBinary, Integer, Sequence, Color, Password,
                            UUID, URL)
from unittest import skipIf

try:
    import cryptography  # noqa
    has_cryptography = True
except:
    has_cryptography = False

try:
    import passlib  # noqa
    has_passlib = True
except:
    has_passlib = False

try:
    import colour  # noqa
    has_colour = True
except:
    has_colour = False

try:
    import furl  # noqa
    has_furl = True
except:
    has_furl = False


Model = Declarations.Model
register = Declarations.register


class OneColumn(Column):
    sqlalchemy_type = SA_Integer


class TestColumn(TestCase):

    def test_forbid_instance(self):
        try:
            Column()
            self.fail("Column mustn't be instanciated")
        except FieldException:
            pass

    def test_without_label(self):
        column = OneColumn()
        column.get_sqlalchemy_mapping(None, None, 'a_column', None)
        self.assertEqual(column.label, 'A column')


def simple_column(ColumnType=None, **kwargs):

    @register(Model)
    class Test:

        id = Integer(primary_key=True)
        col = ColumnType(**kwargs)


def column_with_foreign_key():

    @register(Model)
    class Test:

        name = String(primary_key=True)

    @register(Model)
    class Test2:

        id = Integer(primary_key=True)
        test = String(foreign_key=Model.Test.use('name'))


class TestColumns(DBTestCase):

    def test_column_with_type_in_kwargs(self):
        self.init_registry(
            simple_column, ColumnType=Integer, type_=Integer)

    def test_column_with_db_column_name_in_kwargs(self):
        registry = self.init_registry(simple_column, ColumnType=Integer,
                                      db_column_name='another_name')
        test = registry.Test.insert(col=1)
        self.assertEqual(test.col, 1)
        res = registry.execute('select id from test where another_name=1')
        self.assertEqual(res.fetchone()[0], test.id)

    def test_column_with_foreign_key(self):
        registry = self.init_registry(column_with_foreign_key)
        registry.Test.insert(name='test')
        registry.Test2.insert(test='test')

    def test_integer(self):
        registry = self.init_registry(simple_column, ColumnType=Integer)
        test = registry.Test.insert(col=1)
        self.assertEqual(test.col, 1)

    def test_integer_str_foreign_key(self):
        registry = self.init_registry(
            simple_column, ColumnType=Integer, foreign_key='Model.Test=>id')
        test = registry.Test.insert()
        test2 = registry.Test.insert(col=test.id)
        self.assertEqual(test2.col, test.id)

    def test_big_integer(self):
        registry = self.init_registry(simple_column, ColumnType=BigInteger)
        test = registry.Test.insert(col=1)
        self.assertEqual(test.col, 1)

    def test_small_integer(self):
        registry = self.init_registry(simple_column, ColumnType=SmallInteger)
        test = registry.Test.insert(col=1)
        self.assertEqual(test.col, 1)

    def test_Float(self):
        registry = self.init_registry(simple_column, ColumnType=Float)
        test = registry.Test.insert(col=1.0)
        self.assertEqual(test.col, 1.0)

    def test_decimal(self):
        from decimal import Decimal as D

        registry = self.init_registry(simple_column, ColumnType=Decimal)
        test = registry.Test.insert(col=D('1.0'))
        self.assertEqual(test.col, D('1.0'))

    def test_setter_decimal(self):
        from decimal import Decimal as D

        registry = self.init_registry(simple_column, ColumnType=Decimal)
        test = registry.Test.insert()
        test.col = '1.0'
        self.assertEqual(test.col, D('1.0'))

    def test_boolean(self):
        registry = self.init_registry(simple_column, ColumnType=Boolean)
        test = registry.Test.insert(col=True)
        self.assertEqual(test.col, True)

    def test_boolean_with_default(self):
        registry = self.init_registry(simple_column, ColumnType=Boolean,
                                      default=False)
        test = registry.Test.insert()
        self.assertEqual(test.col, False)

    def test_string(self):
        registry = self.init_registry(simple_column, ColumnType=String)
        test = registry.Test.insert(col='col')
        self.assertEqual(test.col, 'col')

    def test_string_with_False(self):
        registry = self.init_registry(simple_column, ColumnType=String)
        test = registry.Test.insert(col=False)
        self.assertEqual(test.col, False)
        self.registry.flush()
        self.registry.expire(test, ['col'])
        self.assertEqual(test.col, '')

    def test_string_set_False(self):
        registry = self.init_registry(simple_column, ColumnType=String)
        test = registry.Test.insert()
        test.col = False
        self.assertEqual(test.col, False)
        self.registry.flush()
        self.registry.expire(test, ['col'])
        self.assertEqual(test.col, '')

    def test_string_query_False(self):
        registry = self.init_registry(simple_column, ColumnType=String)
        test = registry.Test.insert()
        self.registry.Test.query().filter_by(id=test.id).update({'col': False})
        self.assertEqual(test.col, False)
        self.registry.expire(test, ['col'])
        self.assertEqual(test.col, '')

    @skipIf(not has_cryptography, "cryptography is not installed")
    def test_string_with_encrypt_key(self):
        registry = self.init_registry(simple_column, ColumnType=String,
                                      encrypt_key='secretkey')
        test = registry.Test.insert(col='col')
        self.assertEqual(test.col, 'col')
        res = registry.execute('select col from test where id = %s' % test.id)
        res = res.fetchall()[0][0]
        self.assertNotEqual(res, 'col')

    def test_string_with_size(self):
        registry = self.init_registry(
            simple_column, ColumnType=String, size=100)
        test = registry.Test.insert(col='col')
        self.assertEqual(test.col, 'col')

    @skipIf(not has_passlib, "passlib is not installed")
    def test_password(self):
        registry = self.init_registry(simple_column, ColumnType=Password,
                                      crypt_context={'schemes': ['md5_crypt']})
        test = registry.Test.insert(col='col')
        self.assertEqual(test.col, 'col')
        self.assertNotEqual(
            registry.execute('Select col from test').fetchone()[0], 'col')

    def test_text(self):
        registry = self.init_registry(simple_column, ColumnType=Text)
        test = registry.Test.insert(col='col')
        self.assertEqual(test.col, 'col')

    def test_text_with_False(self):
        registry = self.init_registry(simple_column, ColumnType=Text)
        test = registry.Test.insert(col=False)
        self.assertEqual(test.col, False)
        self.registry.flush()
        self.registry.expire(test, ['col'])
        self.assertEqual(test.col, '')

    def test_text_set_False(self):
        registry = self.init_registry(simple_column, ColumnType=Text)
        test = registry.Test.insert()
        test.col = False
        self.assertEqual(test.col, False)
        self.registry.flush()
        self.registry.expire(test, ['col'])
        self.assertEqual(test.col, '')

    def test_text_query_False(self):
        registry = self.init_registry(simple_column, ColumnType=Text)
        test = registry.Test.insert()
        self.registry.Test.query().filter_by(id=test.id).update({'col': False})
        self.assertEqual(test.col, False)
        self.registry.expire(test, ['col'])
        self.assertEqual(test.col, '')

    def test_ustring(self):
        registry = self.init_registry(simple_column, ColumnType=uString)
        test = registry.Test.insert(col='col')
        self.assertEqual(test.col, 'col')

    def test_ustring_with_size(self):
        registry = self.init_registry(
            simple_column, ColumnType=uString, size=100)
        test = registry.Test.insert(col='col')
        self.assertEqual(test.col, 'col')

    def test_ustring_with_False(self):
        registry = self.init_registry(simple_column, ColumnType=uString)
        test = registry.Test.insert(col=False)
        self.assertEqual(test.col, False)
        self.registry.flush()
        self.registry.expire(test, ['col'])
        self.assertEqual(test.col, '')

    def test_ustring_set_False(self):
        registry = self.init_registry(simple_column, ColumnType=uString)
        test = registry.Test.insert()
        test.col = False
        self.assertEqual(test.col, False)
        self.registry.flush()
        self.registry.expire(test, ['col'])
        self.assertEqual(test.col, '')

    def test_ustring_query_False(self):
        registry = self.init_registry(simple_column, ColumnType=uString)
        test = registry.Test.insert()
        self.registry.Test.query().filter_by(id=test.id).update({'col': False})
        self.assertEqual(test.col, False)
        self.registry.expire(test, ['col'])
        self.assertEqual(test.col, '')

    def test_utext(self):
        registry = self.init_registry(simple_column, ColumnType=uText)
        test = registry.Test.insert(col='col')
        self.assertEqual(test.col, 'col')

    def test_utext_with_False(self):
        registry = self.init_registry(simple_column, ColumnType=uText)
        test = registry.Test.insert(col=False)
        self.assertEqual(test.col, False)
        self.registry.flush()
        self.registry.expire(test, ['col'])
        self.assertEqual(test.col, '')

    def test_utext_set_False(self):
        registry = self.init_registry(simple_column, ColumnType=uText)
        test = registry.Test.insert()
        test.col = False
        self.assertEqual(test.col, False)
        self.registry.flush()
        self.registry.expire(test, ['col'])
        self.assertEqual(test.col, '')

    def test_utext_query_False(self):
        registry = self.init_registry(simple_column, ColumnType=uText)
        test = registry.Test.insert()
        self.registry.Test.query().filter_by(id=test.id).update({'col': False})
        self.assertEqual(test.col, False)
        self.registry.expire(test, ['col'])
        self.assertEqual(test.col, '')

    def test_date(self):
        from datetime import date

        now = date.today()
        registry = self.init_registry(simple_column, ColumnType=Date)
        test = registry.Test.insert(col=now)
        self.assertEqual(test.col, now)

    def test_datetime(self):
        import datetime
        import time
        import pytz

        timezone = pytz.timezone(time.tzname[0])
        now = datetime.datetime.now().replace(tzinfo=timezone)
        registry = self.init_registry(simple_column, ColumnType=DateTime)
        test = registry.Test.insert(col=now)
        self.assertEqual(test.col, now)

    def test_datetime_none_value(self):
        registry = self.init_registry(simple_column, ColumnType=DateTime)
        test = registry.Test.insert(col=None)
        self.assertIsNone(test.col)

    def test_datetime_str_conversion_1(self):
        import datetime
        import time
        import pytz

        timezone = pytz.timezone(time.tzname[0])
        now = datetime.datetime.now().replace(tzinfo=timezone)
        registry = self.init_registry(simple_column, ColumnType=DateTime)
        test = registry.Test.insert(col=now.strftime('%Y-%m-%d %H:%M:%S.%f%z'))
        self.assertEqual(test.col, now)

    def test_datetime_str_conversion_2(self):
        import datetime
        import time
        import pytz

        timezone = pytz.timezone(time.tzname[0])
        now = timezone.localize(datetime.datetime.now())
        registry = self.init_registry(simple_column, ColumnType=DateTime)
        test = registry.Test.insert(col=now.strftime('%Y-%m-%d %H:%M:%S.%f%Z'))
        self.assertEqual(test.col, now)

    def test_datetime_str_conversion_3(self):
        import datetime
        import time
        import pytz

        timezone = pytz.timezone(time.tzname[0])
        now = timezone.localize(datetime.datetime.now())
        registry = self.init_registry(simple_column, ColumnType=DateTime)
        test = registry.Test.insert(col=now.strftime('%Y-%m-%d %H:%M:%S.%f'))
        self.assertEqual(test.col, now)

    def test_datetime_str_conversion_4(self):
        import datetime
        import time
        import pytz

        timezone = pytz.timezone(time.tzname[0])
        now = timezone.localize(datetime.datetime.now())
        registry = self.init_registry(simple_column, ColumnType=DateTime)
        test = registry.Test.insert(col=now.strftime('%Y-%m-%d %H:%M:%S'))
        self.assertEqual(test.col, now.replace(microsecond=0))

    def test_datetime_by_property(self):
        import datetime
        import time
        import pytz

        timezone = pytz.timezone(time.tzname[0])
        now = datetime.datetime.now().replace(tzinfo=timezone)
        registry = self.init_registry(simple_column, ColumnType=DateTime)
        test = registry.Test.insert()
        test.col = now
        self.assertEqual(test.col, now)

    def test_datetime_by_property_none_value(self):
        registry = self.init_registry(simple_column, ColumnType=DateTime)
        test = registry.Test.insert()
        test.col = None
        self.assertIsNone(test.col)

    def test_datetime_str_conversion_1_by_property(self):
        import datetime
        import time
        import pytz

        timezone = pytz.timezone(time.tzname[0])
        now = datetime.datetime.now().replace(tzinfo=timezone)
        registry = self.init_registry(simple_column, ColumnType=DateTime)
        test = registry.Test.insert()
        test.col = now.strftime('%Y-%m-%d %H:%M:%S.%f%z')
        self.assertEqual(test.col, now)

    def test_datetime_str_conversion_2_by_property(self):
        import datetime
        import time
        import pytz

        timezone = pytz.timezone(time.tzname[0])
        now = timezone.localize(datetime.datetime.now())
        registry = self.init_registry(simple_column, ColumnType=DateTime)
        test = registry.Test.insert()
        test.col = now.strftime('%Y-%m-%d %H:%M:%S.%f%Z')
        self.assertEqual(test.col, now)

    def test_datetime_str_conversion_3_by_property(self):
        import datetime
        import time
        import pytz

        timezone = pytz.timezone(time.tzname[0])
        now = timezone.localize(datetime.datetime.now())
        registry = self.init_registry(simple_column, ColumnType=DateTime)
        test = registry.Test.insert()
        test.col = now.strftime('%Y-%m-%d %H:%M:%S.%f')
        self.assertEqual(test.col, now)

    def test_datetime_str_conversion_4_by_property(self):
        import datetime
        import time
        import pytz

        timezone = pytz.timezone(time.tzname[0])
        now = timezone.localize(datetime.datetime.now())
        registry = self.init_registry(simple_column, ColumnType=DateTime)
        test = registry.Test.insert()
        test.col = now.strftime('%Y-%m-%d %H:%M:%S')
        self.assertEqual(test.col, now.replace(microsecond=0))

    def test_datetime_by_query(self):
        import datetime
        import time
        import pytz

        timezone = pytz.timezone(time.tzname[0])
        now = datetime.datetime.now().replace(tzinfo=timezone)
        registry = self.init_registry(simple_column, ColumnType=DateTime)
        test = registry.Test.insert()
        registry.Test.query().update(dict(col=now))
        self.assertEqual(test.col, now)

    def test_datetime_by_query_none_value(self):
        registry = self.init_registry(simple_column, ColumnType=DateTime)
        test = registry.Test.insert()
        registry.Test.query().update(dict(col=None))
        self.assertIsNone(test.col)

    def test_datetime_str_conversion_1_by_query(self):
        import datetime
        import time
        import pytz

        timezone = pytz.timezone(time.tzname[0])
        now = datetime.datetime.now().replace(tzinfo=timezone)
        registry = self.init_registry(simple_column, ColumnType=DateTime)
        test = registry.Test.insert()
        registry.Test.query().update(
            dict(col=now.strftime('%Y-%m-%d %H:%M:%S.%f%z')))
        registry.expire(test, ['col'])
        self.assertEqual(test.col, now)

    def test_datetime_str_conversion_2_by_query(self):
        import datetime
        import time
        import pytz

        timezone = pytz.timezone(time.tzname[0])
        now = timezone.localize(datetime.datetime.now())
        registry = self.init_registry(simple_column, ColumnType=DateTime)
        test = registry.Test.insert()
        registry.Test.query().update(
            dict(col=now.strftime('%Y-%m-%d %H:%M:%S.%f%Z')))
        registry.expire(test, ['col'])
        self.assertEqual(test.col, now)

    def test_datetime_str_conversion_3_by_query(self):
        import datetime
        import time
        import pytz

        timezone = pytz.timezone(time.tzname[0])
        now = timezone.localize(datetime.datetime.now())
        registry = self.init_registry(simple_column, ColumnType=DateTime)
        test = registry.Test.insert()
        registry.Test.query().update(
            dict(col=now.strftime('%Y-%m-%d %H:%M:%S.%f')))
        registry.expire(test, ['col'])
        self.assertEqual(test.col, now)

    def test_datetime_str_conversion_4_by_query(self):
        import datetime
        import time
        import pytz

        timezone = pytz.timezone(time.tzname[0])
        now = timezone.localize(datetime.datetime.now())
        registry = self.init_registry(simple_column, ColumnType=DateTime)
        test = registry.Test.insert()
        registry.Test.query().update(
            dict(col=now.strftime('%Y-%m-%d %H:%M:%S')))
        registry.expire(test, ['col'])
        self.assertEqual(test.col, now.replace(microsecond=0))

    def test_datetime_by_query_filter(self):
        import datetime
        import time
        import pytz

        timezone = pytz.timezone(time.tzname[0])
        now = datetime.datetime.now().replace(tzinfo=timezone)
        registry = self.init_registry(simple_column, ColumnType=DateTime)
        test = registry.Test.insert(col=now)
        Test = registry.Test
        self.assertIs(Test.query().filter(Test.col == now).one(), test)

    def test_datetime_str_conversion_1_by_query_filter(self):
        import datetime
        import time
        import pytz

        timezone = pytz.timezone(time.tzname[0])
        now = datetime.datetime.now().replace(tzinfo=timezone)
        registry = self.init_registry(simple_column, ColumnType=DateTime)
        test = registry.Test.insert(col=now)
        Test = registry.Test
        self.assertIs(
            Test.query().filter(
                Test.col == now.strftime('%Y-%m-%d %H:%M:%S.%f%z')).one(),
            test)

    def test_datetime_str_conversion_2_by_query_filter(self):
        import datetime
        import time
        import pytz

        timezone = pytz.timezone(time.tzname[0])
        now = timezone.localize(datetime.datetime.now())
        registry = self.init_registry(simple_column, ColumnType=DateTime)
        test = registry.Test.insert(col=now)
        Test = registry.Test
        self.assertIs(
            Test.query().filter(
                Test.col == now.strftime('%Y-%m-%d %H:%M:%S.%f%Z')).one(),
            test)

    def test_datetime_str_conversion_3_by_query_filter(self):
        import datetime
        import time
        import pytz

        timezone = pytz.timezone(time.tzname[0])
        now = timezone.localize(datetime.datetime.now())
        registry = self.init_registry(simple_column, ColumnType=DateTime)
        test = registry.Test.insert(col=now)
        Test = registry.Test
        self.assertIs(
            Test.query().filter(
                Test.col == now.strftime('%Y-%m-%d %H:%M:%S.%f')).one(), test)

    def test_interval(self):
        from datetime import timedelta

        dt = timedelta(days=5)
        registry = self.init_registry(simple_column, ColumnType=Interval)
        test = registry.Test.insert(col=dt)
        self.assertEqual(test.col, dt)

    def test_time(self):
        from datetime import time

        now = time()
        registry = self.init_registry(simple_column, ColumnType=Time)
        test = registry.Test.insert(col=now)
        self.assertEqual(test.col, now)

    def test_large_binary(self):
        from os import urandom

        blob = urandom(100000)
        registry = self.init_registry(simple_column, ColumnType=LargeBinary)

        test = registry.Test.insert(col=blob)
        self.assertEqual(test.col, blob)

    def test_selection(self):
        SELECTIONS = [
            ('admin', 'Admin'),
            ('regular-user', 'Regular user')
        ]

        registry = self.init_registry(
            simple_column, ColumnType=Selection, selections=SELECTIONS)
        test = registry.Test.insert(col=SELECTIONS[0][0])
        self.assertEqual(test.col, SELECTIONS[0][0])
        self.assertEqual(test.col.label, SELECTIONS[0][1])
        test = registry.Test.query().first()
        self.assertEqual(test.col, SELECTIONS[0][0])
        self.assertEqual(test.col.label, SELECTIONS[0][1])
        with self.assertRaises(FieldException):
            test.col = 'bad value'

    def test_selection_with_none_value(self):
        SELECTIONS = [
            ('admin', 'Admin'),
            ('regular-user', 'Regular user')
        ]

        registry = self.init_registry(
            simple_column, ColumnType=Selection, selections=SELECTIONS)
        t = registry.Test.insert()
        self.assertIsNone(t.col)

    @skipIf(True, "Find a better way to check a column selection by query "
                  "update")
    def test_selection_change_by_query(self):
        SELECTIONS = [
            ('admin', 'Admin'),
            ('regular-user', 'Regular user')
        ]

        registry = self.init_registry(
            simple_column, ColumnType=Selection, selections=SELECTIONS)
        registry.Test.insert(col=SELECTIONS[0][0])
        with self.assertRaises(StatementError):
            registry.Test.query().update({'col': 'bad value'})

    def test_selection_like_comparator(self):
        SELECTIONS = [
            ('admin', 'Admin'),
            ('regular-user', 'Regular user')
        ]

        registry = self.init_registry(
            simple_column, ColumnType=Selection, selections=SELECTIONS)
        Test = registry.Test
        t = Test.insert(col=SELECTIONS[0][0])
        t1 = Test.query().filter(Test.col.like('%admin%')).one()
        self.assertIs(t, t1)

    def test_selection_key_other_than_str(self):
        SELECTIONS = [
            (0, 'Admin'),
            (1, 'Regular user')
        ]

        with self.assertRaises(FieldException):
            self.init_registry(
                simple_column, ColumnType=Selection, selections=SELECTIONS)

    def test_selection_comparator(self):
        SELECTIONS = [
            ('admin', 'Admin'),
            ('regular-user', 'Regular user')
        ]

        registry = self.init_registry(
            simple_column, ColumnType=Selection, selections=SELECTIONS)
        registry.Test.insert(col=SELECTIONS[0][0])
        registry.Test.query().filter(
            registry.Test.col.in_(['admin', 'regular-user'])).first()

    def test_selection_use_method(self):
        SELECTIONS = [
            ('admin', 'Admin'),
            ('regular-user', 'Regular user')
        ]

        def add_selection():

            @register(Model)
            class Test:

                id = Integer(primary_key=True)
                col = Selection(selections='get_selection')

                @classmethod
                def get_selection(cls):
                    return SELECTIONS

        registry = self.init_registry(add_selection)
        registry.Test.insert(col=SELECTIONS[0][0])
        registry.Test.query().filter(
            registry.Test.col.in_(['admin', 'regular-user'])).first()

    def test_json(self):
        registry = self.init_registry(simple_column, ColumnType=Json)
        val = {'a': 'Test'}
        test = registry.Test.insert(col=val)
        self.assertEqual(test.col, val)

    def test_json_update(self):
        registry = self.init_registry(simple_column, ColumnType=Json)
        test = registry.Test.insert(col={'a': 'test'})
        test.col['b'] = 'test'
        self.assertEqual(test.col, {'a': 'test', 'b': 'test'})

    def test_json_simple_filter(self):
        registry = self.init_registry(simple_column, ColumnType=Json)
        Test = registry.Test
        Test.insert(col={'a': 'test'})
        Test.insert(col={'a': 'test'})
        Test.insert(col={'b': 'test'})
        self.assertEqual(
            Test.query().filter(Test.col == {'a': 'test'}).count(), 2)

    def test_json_null(self):
        registry = self.init_registry(simple_column, ColumnType=Json)
        Test = registry.Test
        Test.insert(col=None)
        Test.insert(col=None)
        Test.insert(col={'a': 'test'})
        self.assertEqual(Test.query().filter(Test.col == Json.Null).count(), 2)
        self.assertEqual(Test.query().filter(Test.col != Json.Null).count(), 1)

    def test_add_default_classmethod(self):
        val = 'test'

        def add_in_registry():

            @register(Model)
            class Test:

                id = Integer(primary_key=True)
                val = String(default="get_val")

                @classmethod
                def get_val(cls):
                    return val

        registry = self.init_registry(add_in_registry)
        t = registry.Test.insert()
        self.assertEqual(t.val, val)

    def test_add_default_without_classmethod(self):
        value = 'test'

        def add_in_registry():

            @register(Model)
            class Test:

                id = Integer(primary_key=True)
                val = String(default="get_val")

                def get_val(cls):
                    return value

        registry = self.init_registry(add_in_registry)
        t = registry.Test.insert()
        self.assertEqual(t.val, "get_val")

    def test_add_default_by_var(self):
        value = 'test'

        def add_in_registry():

            @register(Model)
            class Test:

                id = Integer(primary_key=True)
                val = String(default=value)

        registry = self.init_registry(add_in_registry)
        t = registry.Test.insert()
        self.assertEqual(t.val, value)

    def test_add_default(self):

        def add_in_registry():

            @register(Model)
            class Test:

                id = Integer(primary_key=True)
                val = String(default='value')

        registry = self.init_registry(add_in_registry)
        t = registry.Test.insert()
        self.assertEqual(t.val, 'value')

    def test_add_field_as_default(self):

        def add_in_registry():

            @register(Model)
            class Test:

                id = Integer(primary_key=True)
                val = String(default='val')

        registry = self.init_registry(add_in_registry)
        t = registry.Test.insert()
        self.assertEqual(t.val, 'val')

    def test_sequence(self):
        registry = self.init_registry(simple_column, ColumnType=Sequence)
        self.assertEqual(registry.Test.insert().col, "1")
        self.assertEqual(registry.Test.insert().col, "2")
        self.assertEqual(registry.Test.insert().col, "3")
        self.assertEqual(registry.Test.insert().col, "4")
        Seq = registry.System.Sequence
        self.assertEqual(
            Seq.query().filter(Seq.code == 'Model.Test=>col').count(), 1)

    def test_sequence_with_code_and_formater(self):
        registry = self.init_registry(simple_column, ColumnType=Sequence,
                                      code="SO", formater="{code}-{seq:06d}")
        self.assertEqual(registry.Test.insert().col, "SO-000001")
        self.assertEqual(registry.Test.insert().col, "SO-000002")
        self.assertEqual(registry.Test.insert().col, "SO-000003")
        self.assertEqual(registry.Test.insert().col, "SO-000004")
        Seq = registry.System.Sequence
        self.assertEqual(Seq.query().filter(Seq.code == 'SO').count(), 1)

    def test_sequence_with_foreign_key(self):
        with self.assertRaises(FieldException):
            self.init_registry(simple_column, ColumnType=Sequence,
                               foreign_key=Model.System.Model.use('name'))

    def test_sequence_with_default(self):
        with self.assertRaises(FieldException):
            self.init_registry(simple_column, ColumnType=Sequence,
                               default='default value')

    @skipIf(not has_colour, "colour is not installed")
    def test_color(self):
        color = '#F5F5F5'
        registry = self.init_registry(simple_column, ColumnType=Color)
        test = registry.Test.insert(col=color)
        self.assertEqual(test.col.hex, colour.Color(color).hex)

    @skipIf(not has_colour, "colour is not installed")
    def test_setter_color(self):
        color = '#F5F5F5'
        registry = self.init_registry(simple_column, ColumnType=Color)
        test = registry.Test.insert()
        test.col = color
        self.assertEqual(test.col.hex, colour.Color(color).hex)

    def test_uuid_binary_1(self):
        from uuid import uuid1
        uuid = uuid1()
        registry = self.init_registry(simple_column, ColumnType=UUID)
        test = registry.Test.insert(col=uuid)
        self.assertIs(test.col, uuid)

    def test_uuid_binary_3(self):
        from uuid import uuid3, NAMESPACE_DNS
        uuid = uuid3(NAMESPACE_DNS, 'python.org')
        registry = self.init_registry(simple_column, ColumnType=UUID)
        test = registry.Test.insert(col=uuid)
        self.assertIs(test.col, uuid)

    def test_uuid_binary_4(self):
        from uuid import uuid4
        uuid = uuid4()
        registry = self.init_registry(simple_column, ColumnType=UUID)
        test = registry.Test.insert(col=uuid)
        self.assertIs(test.col, uuid)

    def test_uuid_binary_5(self):
        from uuid import uuid5, NAMESPACE_DNS
        uuid = uuid5(NAMESPACE_DNS, 'python.org')
        registry = self.init_registry(simple_column, ColumnType=UUID)
        test = registry.Test.insert(col=uuid)
        self.assertIs(test.col, uuid)

    def test_uuid_char32(self):
        from uuid import uuid1
        uuid = uuid1()
        registry = self.init_registry(simple_column, ColumnType=UUID,
                                      binary=False)
        test = registry.Test.insert(col=uuid)
        self.assertIs(test.col, uuid)

    @skipIf(not has_furl, "furl is not installed")
    def test_URL(self):
        f = furl.furl('http://doc.anyblok.org')
        registry = self.init_registry(simple_column, ColumnType=URL)
        test = registry.Test.insert(col=f)
        self.assertEqual(test.col, f)

    @skipIf(not has_furl, "furl is not installed")
    def test_setter_URL(self):
        f = 'http://doc.anyblok.org'
        registry = self.init_registry(simple_column, ColumnType=URL)
        test = registry.Test.insert()
        test.col = f
        self.assertEqual(test.col.url, f)

    @skipIf(not has_furl, "furl is not installed")
    def test_URL_2(self):
        f = 'http://doc.anyblok.org'
        registry = self.init_registry(simple_column, ColumnType=URL)
        test = registry.Test.insert(col=f)
        self.assertEqual(test.col.url, f)

    @skipIf(not has_furl, "furl is not installed")
    def test_URL_3(self):
        f = 'http://doc.anyblok.org'
        registry = self.init_registry(simple_column, ColumnType=URL)
        registry.Test.insert(col=f)
        Test = registry.Test
        test = Test.query().filter(Test.col == f).one()
        self.assertEqual(test.col.url, f)

    @skipIf(not has_furl, "furl is not installed")
    def test_URL_4(self):
        f = furl.furl('http://doc.anyblok.org')
        registry = self.init_registry(simple_column, ColumnType=URL)
        registry.Test.insert(col=f)
        Test = registry.Test
        test = Test.query().filter(Test.col == f).one()
        self.assertEqual(test.col.url, f.url)



# This file is a part of the AnyBlok project
#
#    Copyright (C) 2015 Jean-Sebastien SUZANNE <jssuzanne@anybox.fr>
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file,You can
# obtain one at http://mozilla.org/MPL/2.0/.
from anyblok import Declarations
from lxml import etree
from .exceptions import XMLImporterException


register = Declarations.register
IO = Declarations.Model.IO
if_exist = 'overwrite'
if_does_not_exist = 'create'
on_error = 'ignore'


@register(IO)
class Importer:

    @classmethod
    def get_mode_choices(cls):
        res = super(Importer, cls).get_mode_choices()
        res.update({'Model.IO.Importer.XML': 'XML'})
        return res


@register(IO.Importer)
class XML:

    def __init__(self, importer, blokname=None):
        self.importer = importer
        self.error_found = []
        self.created_entries = []
        self.updated_entries = []
        self.params = {}
        self.two_way_external_id = {}
        self.blokname = blokname

    def commit(self):
        if self.error_found:
            return False

        self.importer.commit()
        return True

    def _raise(self, msg, on_error=on_error, **kwargs):
        self.error_found.append(str(msg))
        if on_error == 'raise':
            raise XMLImporterException(msg)

    def find_entry(self, model=None, external_id=None, param=None, **kwargs):
        if (model, param) in self.params:
            return self.params[(model, param)]
        if model and external_id:
            if (model, external_id) in self.two_way_external_id:
                return self.two_way_external_id[(model, external_id)]

            entry = self.registry.IO.Mapping.get(model, external_id)
            if entry:
                return entry

        return None

    def check_entry_before_import(self, record, entry, if_exist=if_exist,
                                  if_does_not_exist=if_does_not_exist,
                                  **kwargs):
        if entry:
            if if_exist == 'pass':
                return True, entry
            elif if_exist == 'raise':
                self._raise("The tag %r (%r) already exist" % (
                    record.tag, record.attrib), **kwargs)
                # case of raise at the end, must not bloc import
                return True, None
        else:
            if if_does_not_exist == 'pass':
                return True, None
            elif if_does_not_exist == 'raise':
                self._raise("The tag %r (%r) does not exist" % (
                    record.tag, record.attrib), **kwargs)
                # case of raise at the end, must not bloc import
                return True, None

        return False, entry

    def update_x2M(self, entry, inValues, notInValues):
        for field in inValues.keys():
            if field in notInValues:
                continue

            # temporaly deactivate auto flush to let time
            # orm to fill the remote column with foreign_key
            # which should be not null
            with self.registry.session.no_autoflush:
                getattr(entry, field).extend(inValues[field])

    def create_entry(self, Model, values, two_way, **kwargs):
        try:
            insert_values = {x: y for x, y in values.items()
                             if not isinstance(y, list)}
            if two_way:
                return_entry = Model(**insert_values)
            else:
                return_entry = Model.insert(**insert_values)

            self.update_x2M(return_entry, values, insert_values)
            self.created_entries.append(return_entry)
            return return_entry
        except Exception as e:
            self._raise(e, **kwargs)

    def map_imported_entry(self, model, param, external_id, two_way, entry,
                           if_exist=if_exist, **kwargs):
        if entry:
            if model and param:
                if (model, param) not in self.params:
                    self.params[(model, param)] = entry
                else:
                    if self.params[(model, param)] != entry:
                        self._raise("Overwrite the params (%r, %r)" % (
                            model, param), **kwargs)

            if external_id:
                if two_way:
                    self.two_way_external_id[(
                        model, external_id)] = entry
                else:
                    raiseifexist = if_exist != 'overwrite'
                    self.registry.IO.Mapping.set(external_id, entry,
                                                 blokname=self.blokname,
                                                 raiseifexist=raiseifexist)

    def import_entry(self, entry, values, model=None, external_id=None,
                     param=None, if_exist=if_exist,
                     if_does_not_exist=if_does_not_exist,
                     two_way=False, **kwargs):
        Model = self.registry.get(model)
        return_entry = entry
        if entry:
            if if_exist == 'continue':
                return_entry = entry
            elif if_exist == 'create':
                return_entry = self.create_entry(
                    Model, values, two_way, **kwargs)
            elif if_exist == 'overwrite':
                try:
                    insert_values = {x: y for x, y in values.items()
                                     if not isinstance(y, list)}
                    if insert_values:
                        entry.update(**insert_values)

                    self.update_x2M(entry, values, insert_values)
                    self.updated_entries.append(entry)
                except Exception as e:
                    self._raise(e, **kwargs)
        elif if_does_not_exist == 'create':
                return_entry = self.create_entry(
                    Model, values, two_way, **kwargs)

        self.map_imported_entry(
            model, param, external_id, two_way, return_entry,
            if_exist=if_exist, **kwargs)

        return return_entry

    def import_multi_values(self, records, model):
        vals = []
        for record in records:
            val = self.import_record(
                record,
                model=record.attrib.get('model', model),
                two_way=True
            )
            if not val:
                continue

            vals.append(val)

        return vals

    def import_value(self, record, ctype, model, on_error=on_error):
        val = self.import_record(record, model=model)
        if ctype in ('One2One', 'Many2One'):
            return val
        else:
            if not val:
                return None

            pks = val.to_primary_keys()
            vals = [x for x in pks.values()]
            if len(vals) > 1:
                self._raise(
                    "Multi foreign key for %r (%r) are not "
                    "implemented" % (record.tag, record.attrib),
                    on_error=on_error)

            return vals[0]

    def get_from_param(self, param, model):
        if param:
            if model:
                if (model, param) in self.params:
                    return self.params[(model, param)]

            else:
                self._raise('Param %r waiting for None model ' % param,
                            on_error=on_error)

        return None

    def import_field(self, field, ctype, model=None, on_error=on_error):
        model = field.attrib.get('model', model)
        param = field.attrib.get('param')

        res = self.get_from_param(param, model)
        if res:
            return res

        val = field.text
        external_id = False
        if 'external_id' in field.attrib:
            external_id = True
            val = field.attrib['external_id']

        try:
            res = None
            if ctype in ('Many2One', 'One2One') and external_id:
                if model and (model, val) in self.two_way_external_id:
                    res = self.two_way_external_id[(model, val)]

            if not res:
                res = self.importer.str2value(val, ctype,
                                              external_id=external_id,
                                              model=model)
            if param:
                self.params[(model, param)] = res

            return res
        except Exception as e:
            self._raise(e, on_error=on_error)
            return None

    def import_record(self, record, two_way=False, **kwargs):
        kwargs.update(record.attrib)
        _kw = {}
        if 'on_error' in kwargs:
            _kw['on_error'] = kwargs['on_error']

        entry = self.find_entry(**kwargs)
        mustbereturned, entry = self.check_entry_before_import(
            record, entry, **kwargs)

        if mustbereturned:
            return entry

        if 'model' not in kwargs:
            self._raise("No model found for record %r (%r)" % (
                record.tag, kwargs), **_kw)
            return None

        Model = self.registry.get(kwargs['model'])
        fields_description = Model.fields_description()
        values = {}
        for field in record.getchildren():
            if 'on_error' in field.attrib:
                _kw['on_error'] = field.attrib['on_error']

            if not self.validate_field(field, Model, fields_description,
                                       **_kw):
                continue

            field_name = field.attrib['name']
            field_model = fields_description[field_name]['model']
            field_type = fields_description[field_name]['type']
            val = self._import_record(field, field_model, field_type, **_kw)
            values[field_name] = val

        entry = self.import_entry(entry, values, two_way=two_way, **kwargs)
        if not two_way and self.two_way_external_id:
            self.two_ways_to_external_id()

        return entry

    def two_ways_to_external_id(self):
        for k, entry in self.two_way_external_id.items():
            model, external_id = k
            # FIXME no controle of if_exist
            self.map_imported_entry(model, None, external_id, False, entry)

    def validate_field(self, field, Model, fields_description, **_kw):
        if field.tag is etree.Comment:
            return False

        if field.tag.lower() != 'field':
            self._raise("Waitting 'field' node, not %r" % field.tag, **_kw)
            return False

        if 'name' not in field.attrib:
            self._raise("field %r (%r) have not attribute name" % (
                field.tag, field.attrib), **_kw)
            return False

        field_name = field.attrib['name']
        if field_name not in fields_description:
            self._raise('Model %r have not field %r' % (
                Model.__registry_name__, field_name))
            return False

        return True

    def _import_record(self, field, field_model, field_type, **_kw):
        children = field.getchildren()

        if children:
            if field_type in ('One2Many', 'Many2Many'):
                vals = self.import_multi_values(children, field_model)
                if vals:
                    return vals

                return []
            else:
                return self.import_value(
                    field, field_type, field_model, **_kw)

        else:
            return self.import_field(
                field, field_type, model=field_model, **_kw)

        return None

    def import_records(self, records):
        for record in records.getchildren():
            if record.tag is etree.Comment:
                continue
            elif record.tag.lower() == 'record':
                self.import_record(record, model=self.importer.model)
            elif record.tag.lower() == 'commit':
                self.commit()
            else:
                self._raise('%r is not known' % record.tag, **records.attrib)

    def run(self):
        records = etree.fromstring(self.importer.file_to_import)
        if records.tag.lower() == 'records':
            self.import_records(records)
        else:
            self._raise('%r is not known' % records.tag)

        if self.error_found:
            if records.attrib.get('on_error', 'raise') == 'raise':
                msg = "Exception found : \n %s" % '\n'.join(self.error_found)
                raise XMLImporterException(msg)

        return {
            'error_found': self.error_found,
            'created_entries': self.created_entries,
            'updated_entries': self.updated_entries,
        }

    @classmethod
    def insert(cls, delimiter=None, quotechar=None, **kwargs):
        kwargs['mode'] = cls.__registry_name__
        if 'model' not in kwargs:
            raise XMLImporterException(
                "The column 'model' is required")

        if not isinstance(kwargs['model'], str):
            kwargs['model'] = kwargs['model'].__registry_name__

        return cls.registry.IO.Importer.insert(**kwargs)



# This file is a part of the AnyBlok project
#
#    Copyright (C) 2014 Jean-Sebastien SUZANNE <jssuzanne@anybox.fr>
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file,You can
# obtain one at http://mozilla.org/MPL/2.0/.
from anyblok.tests.testcase import DBTestCase
from anyblok import Declarations
from anyblok.column import Integer, String
from anyblok.relationship import Many2One

register = Declarations.register
Model = Declarations.Model
Mixin = Declarations.Mixin
Core = Declarations.Core


def simple_subclass_model():

    @register(Model)
    class Test:
        id = Integer(primary_key=True)
        name = String()

    @register(Model)  # noqa
    class Test:
        other = String()


def simple_subclass_Core_Base():

    @register(Core)
    class Base:

        def mymethod(self):
            return 1

    @register(Core)  # noqa
    class Base:

        def mymethod(self):
            return 10 * super(Base, self).mymethod()

    @register(Model)
    class Test:
        pass


def simple_subclass_Core_SqlBase():

    @register(Core)
    class SqlBase:

        def mymethod(self):
            return 1

    @register(Core)  # noqa
    class SqlBase:

        def mymethod(self):
            return 10 * super(SqlBase, self).mymethod()

    @register(Model)
    class Test:
        id = Integer(primary_key=True)


def simple_subclass_model_change_type():

    @register(Model)
    class Test:
        id = Integer(primary_key=True)
        name = String()

    @register(Model)  # noqa
    class Test:
        name = Integer()


def simple_subclass_model_change_type_and_subclass_add_field():

    @register(Model)
    class Test:
        id = Integer(primary_key=True)
        name = String()

    @register(Model)  # noqa
    class Test:
        name = Integer()

    @register(Model)  # noqa
    class Test:
        other = String()


def mixin_one_model():

    @register(Mixin)
    class MixinName:

        name = String()

    @register(Model)
    class Test(Mixin.MixinName):

        id = Integer(primary_key=True)
        other = String()


def mixin_two_model():

    @register(Mixin)
    class MixinName:

        name = String()

    @register(Model)
    class Test(Mixin.MixinName):

        id = Integer(primary_key=True)
        other = String()

    @register(Model)
    class Test2(Mixin.MixinName):

        id = Integer(primary_key=True)
        other = String()


def mixin_one_model_with_subclass():

    @register(Mixin)
    class MixinName:

        name = String()

    @register(Model)
    class Test(Mixin.MixinName):

        id = Integer(primary_key=True)

    @register(Model)  # noqa
    class Test:

        other = String()


def mixin_one_model_by_subclass():

    @register(Mixin)
    class MixinName:

        name = String()

    @register(Model)
    class Test:

        id = Integer(primary_key=True)

    @register(Model)  # noqa
    class Test(Mixin.MixinName):

        other = String()


def mixin_with_foreign_key_one_model():

    @register(Model)
    class TestFk:

        name = String(primary_key=True)

    @register(Mixin)
    class MixinName:

        name = String(foreign_key=Model.TestFk.use('name'))

    @register(Model)
    class Test(Mixin.MixinName):

        id = Integer(primary_key=True)


def mixin_with_foreign_key_two_model():

    @register(Model)
    class TestFk:

        name = String(primary_key=True)

    @register(Mixin)
    class MixinName:

        name = String(foreign_key=Model.TestFk.use('name'))

    @register(Model)
    class Test(Mixin.MixinName):

        id = Integer(primary_key=True)

    @register(Model)
    class Test2(Mixin.MixinName):

        id = Integer(primary_key=True)


def mixin_one_model_by_subclass_and_with():

    @register(Mixin)
    class MixinName:

        name = String()

    @register(Model)
    class Test:

        id = Integer(primary_key=True)

    @register(Model)  # noqa
    class Test(Mixin.MixinName):
        pass

    @register(Model)  # noqa
    class Test:

        other = String()


def mixin_one_model_with_subclass_and_subclass_mixin():

    @register(Mixin)
    class MixinName:

        name = String()

    @register(Model)
    class Test:

        id = Integer(primary_key=True)

    @register(Model)  # noqa
    class Test(Mixin.MixinName):
        pass

    @register(Mixin)  # noqa
    class MixinName:

        other = String()


def inherit_by_another_model():

    @register(Model)
    class MainModel:

        id = Integer(primary_key=True)
        name = String()

    @register(Model)
    class Test(Model.MainModel):

        id = Integer(primary_key=True)
        mainmodel = Integer(foreign_key=Model.MainModel.use('id'))
        other = String()


def inherit_by_two_another_model():

    @register(Model)
    class MainModel:

        id = Integer(primary_key=True)
        name = String()

    @register(Model)
    class Test(Model.MainModel):

        id = Integer(primary_key=True)
        mainmodel = Integer(foreign_key=Model.MainModel.use('id'))
        other = String()

    @register(Model)
    class Test2(Model.MainModel):

        id = Integer(primary_key=True)
        mainmodel = Integer(foreign_key=Model.MainModel.use('id'))
        other = String()


def inherit_by_another_model_and_subclass_mainmodel():

    @register(Model)
    class MainModel:

        id = Integer(primary_key=True)
        name = String()

    @register(Model)
    class Test(Model.MainModel):

        id = Integer(primary_key=True)
        mainmodel = Integer(foreign_key=Model.MainModel.use('id'))

    @register(Model)  # noqa
    class MainModel:

        other = String()


def inherit_base_and_add_method():

    @register(Core)
    class Base:

        def method_from_base(self, x):
            return x * 2

    @register(Model)
    class Test:
        pass


def inherit_base_and_add_method_after_create_model():

    @register(Model)
    class Test:
        pass

    @register(Core)
    class Base:

        def method_from_base(self, x):
            return x * 2


def inherit_base_and_add_method_sub_classes():

    @register(Core)
    class Base:

        def method_from_base(self, x):
            return x * 2

    @register(Model)
    class Test:

        def method_from_base(self, x):
            return super(Test, self).method_from_base(x) + 3


def inherit_base_and_add_method_sub_classes_by_mixin():

    @register(Core)
    class Base:

        def method_from_base(self, x):
            return x * 2

    @register(Mixin)
    class TestMixin:

        def method_from_base(self, x):
            return super(TestMixin, self).method_from_base(x) + 3

    @register(Model)
    class Test(Mixin.TestMixin):
        pass


def inherit_sql_base_on_simple_model():

    @register(Core)
    class Base:

        def is_sql_base(self):
            return False

    @register(Core)
    class SqlBase:

        def is_sql_base(self):
            return True

    @register(Model)
    class Test:
        pass


def inherit_sql_base_on_sql_model():

    @register(Core)
    class Base:

        def is_sql_base(self):
            return False

    @register(Core)
    class SqlBase:

        def is_sql_base(self):
            return True

    @register(Model)
    class Test:

        id = Integer(primary_key=True)


class TestInherit(DBTestCase):

    def check_registry(self, Model):
        t = Model.insert(name="test", other="other")
        t2 = Model.query().first()
        self.assertEqual(t2, t)

    def test_simple_subclass_model(self):
        registry = self.init_registry(simple_subclass_model)
        self.check_registry(registry.Test)

    def test_inherit_multi_mixins(self):
        def add_in_registry():
            Integer = Declarations.Column.Integer
            String = Declarations.Column.String

            @register(Mixin)
            class MixinName:

                name = String()

            @register(Mixin)
            class MixinOther:

                other = String()

            @register(Model)
            class Test(Mixin.MixinName, Mixin.MixinOther):

                id = Integer(primary_key=True)

        registry = self.init_registry(simple_subclass_model)
        self.check_registry(registry.Test)

    def test_inherit_cascade_mixins(self):
        def add_in_registry():
            Integer = Declarations.Column.Integer
            String = Declarations.Column.String

            @register(Mixin)
            class MixinName:

                name = String()

            @register(Mixin)
            class MixinOther:

                other = String()

            @register(Mixin)
            class MTest(Mixin.MixinName, Mixin.MixinOther):
                pass

            @register(Model)
            class Test(Mixin.MTest):

                id = Integer(primary_key=True)

        registry = self.init_registry(simple_subclass_model)
        self.check_registry(registry.Test)

    def test_simple_subclass_Core_Base(self):
        registry = self.init_registry(simple_subclass_Core_Base)
        m = registry.Test()
        self.assertEqual(m.mymethod(), 10)

    def test_simple_subclass_Core_SqlBase(self):
        registry = self.init_registry(simple_subclass_Core_SqlBase)
        m = registry.Test()
        self.assertEqual(m.mymethod(), 10)

    def test_simple_subclass_model_change_type(self):
        registry = self.init_registry(simple_subclass_model_change_type)

        t = registry.Test.insert(name=1)
        t2 = registry.Test.query().first()
        self.assertEqual(t2, t)
        self.assertEqual(t.name, 1)

    def test_simple_subclass_model_change_type_and_subclass_add_field(self):
        registry = self.init_registry(
            simple_subclass_model_change_type_and_subclass_add_field)

        t = registry.Test.insert(name=1, other='other')
        t2 = registry.Test.query().first()
        self.assertEqual(t2, t)
        self.assertEqual(t.name, 1)

    def test_mixin_one_model(self):
        registry = self.init_registry(mixin_one_model)
        self.check_registry(registry.Test)

    def test_mixin_two_model(self):
        registry = self.init_registry(mixin_two_model)
        self.check_registry(registry.Test)
        self.check_registry(registry.Test2)

    def test_mixin_one_model_with_subclass(self):
        registry = self.init_registry(mixin_one_model_with_subclass)
        self.check_registry(registry.Test)

    def test_mixin_one_model_by_subclass(self):
        registry = self.init_registry(mixin_one_model_by_subclass)
        self.check_registry(registry.Test)

    def test_mixin_with_foreign_key_one_model(self):
        registry = self.init_registry(mixin_with_foreign_key_one_model)

        val = 'test'
        registry.TestFk.insert(name=val)
        t = registry.Test.insert(name=val)
        t2 = registry.Test.query().first()
        self.assertEqual(t2, t)

    def test_mixin_with_foreign_key_two_model(self):
        registry = self.init_registry(mixin_with_foreign_key_two_model)

        for val in ('test', 'test2'):
            registry.TestFk.insert(name=val)

        t = registry.Test.insert(name='test')
        t2 = registry.Test.query().first()
        self.assertEqual(t2, t)

        t3 = registry.Test2.insert(name='test2')
        t4 = registry.Test2.query().first()
        self.assertEqual(t3, t4)

    def test_mixin_one_model_by_subclass_and_with(self):
        registry = self.init_registry(mixin_one_model_by_subclass_and_with)
        self.check_registry(registry.Test)

    def test_mixin_one_model_with_subclass_and_subclass_mixin(self):
        registry = self.init_registry(
            mixin_one_model_with_subclass_and_subclass_mixin)
        self.check_registry(registry.Test)

    def test_inherit_by_another_model(self):
        registry = self.init_registry(inherit_by_another_model)
        self.check_registry(registry.Test)

    def test_inherit_by_two_another_model(self):
        registry = self.init_registry(inherit_by_two_another_model)
        self.check_registry(registry.Test)
        self.check_registry(registry.Test2)

    def test_inherit_by_another_model_and_subclass_mainmodel(self):
        registry = self.init_registry(
            inherit_by_another_model_and_subclass_mainmodel)
        self.check_registry(registry.Test)

    def check_inherit_base(self, function, value):
        registry = self.init_registry(function)
        test = registry.Test()
        self.assertEqual(test.method_from_base(2), value)

    def test_inherit_base_and_add_method(self):
        self.check_inherit_base(inherit_base_and_add_method, 4)

    def test_inherit_base_and_add_method_after_create_model(self):
        self.check_inherit_base(inherit_base_and_add_method_after_create_model,
                                4)

    def test_inherit_base_and_add_method_sub_classes(self):
        self.check_inherit_base(inherit_base_and_add_method_sub_classes, 7)

    def test_inherit_base_and_add_method_sub_classes_by_mixin(self):
        self.check_inherit_base(
            inherit_base_and_add_method_sub_classes_by_mixin, 7)

    def check_inherit_sql_base(self, function, value):
        registry = self.init_registry(function)
        test = registry.Test()
        self.assertEqual(test.is_sql_base(), value)

    def test_inherit_sql_base_on_simple_model(self):
        self.check_inherit_sql_base(inherit_sql_base_on_simple_model, False)

    def test_inherit_sql_base_on_sql_model(self):
        self.check_inherit_sql_base(inherit_sql_base_on_sql_model, True)

    def test_inherit_model_with_relation_ship_auto_generate_column(self):

        def add_in_registry():

            @register(Model)
            class Test:
                id = Integer(primary_key=True)

            @register(Model)
            class Test2:
                id = Integer(primary_key=True)
                test = Many2One(model=Model.Test, one2many='test2')

            @register(Model)
            class Test3(Model.Test2):
                id = Integer(primary_key=True)
                test2 = Integer(foreign_key=Model.Test2.use('id'))

        registry = self.init_registry(add_in_registry)
        t = registry.Test.insert()
        t3 = registry.Test3.insert(test=t)
        t2 = registry.Test2.query().one()
        self.assertEqual(t3.test2, t2.id)
        self.assertEqual(t2.test, t)

    def test_inherit_model_with_relation_ship(self):

        def add_in_registry():

            @register(Model)
            class Test:
                id = Integer(primary_key=True)

            @register(Model)
            class Test2:
                id = Integer(primary_key=True)
                test_id = Integer(foreign_key=Model.Test.use('id'))
                test = Many2One(model=Model.Test, one2many='test2')

            @register(Model)
            class Test3(Model.Test2):
                id = Integer(primary_key=True)
                test2 = Integer(foreign_key=Model.Test2.use('id'))

        registry = self.init_registry(add_in_registry)
        t = registry.Test.insert()
        t3 = registry.Test3.insert(test=t)
        t2 = registry.Test2.query().one()
        self.assertEqual(t3.test2, t2.id)
        self.assertEqual(t2.test, t)



# -*- coding: utf-8 -*-
# This file is a part of the AnyBlok project
#
#    Copyright (C) 2014 Jean-Sebastien SUZANNE <jssuzanne@anybox.fr>
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file,You can
# obtain one at http://mozilla.org/MPL/2.0/.
from pkg_resources import iter_entry_points
from anyblok.imp import ImportManager
from .logging import log
from anyblok.environment import EnvironmentManager
from time import sleep
from sys import modules
from os.path import dirname
from logging import getLogger
from os.path import join

logger = getLogger(__name__)


class BlokManagerException(LookupError):
    """ Simple exception to BlokManager """

    def __init__(self, *args, **kwargs):
        EnvironmentManager.set('current_blok', None)
        super(BlokManagerException, self).__init__(*args, **kwargs)


class BlokManager:
    """ Manage the bloks for one process

    A blok has a `setuptools` entrypoint, this entry point is defined
    by the ``entry_points`` attribute in the first load

    The ``bloks`` attribute is a dict with all the loaded entry points

    Use this class to import all the bloks in the entrypoint::

        BlokManager.load()

    """

    bloks = {}
    entry_points = None
    ordered_bloks = []
    auto_install = []
    importers = {}

    @classmethod
    def list(cls):
        """ Return the ordered bloks

        :rtype: list of blok name ordered by loading
        """
        return cls.ordered_bloks

    @classmethod
    def has(cls, blok):
        """ Return True if the blok is loaded

        :param blok: blok name
        :rtype: bool
        """
        return blok and blok in cls.ordered_bloks or False

    @classmethod
    def get(cls, blok):
        """ Return the loaded blok

        :param blok: blok name
        :rtype: blok instance
        :exception: BlokManagerException
        """
        if not cls.has(blok):
            raise BlokManagerException('%r not found' % blok)

        return cls.bloks[blok]

    @classmethod
    def set(cls, blokname, blok):
        """ Add a new blok

        :param blokname: blok name
        :param blok: blok instance
        :exception: BlokManagerException
        """
        if cls.has(blokname):
            raise BlokManagerException('%r already present' % blokname)

        cls.bloks[blokname] = blok
        cls.ordered_bloks.append(blokname)

    @classmethod
    @log(logger)
    def reload(cls):
        """ Reload the entry points

        Empty the ``bloks`` dict and use the ``entry_points`` attribute to
        load bloks
        :exception: BlokManagerException
        """
        if cls.entry_points is None:
            raise BlokManagerException(
                """You must use the ``load`` classmethod before using """
                """``reload``""")

        entry_points = []
        entry_points += cls.entry_points
        cls.unload()
        cls.load(entry_points=entry_points)

    @classmethod
    @log(logger)
    def unload(cls):
        """ Unload all the bloks but not the registry """
        cls.bloks = {}
        cls.ordered_bloks = []
        cls.entry_points = None
        cls.auto_install = []
        from .registry import RegistryManager
        RegistryManager.unload()

    @classmethod
    def get_need_blok_linked_bloks(cls, blok):
        for required in cls.bloks[blok].required:
            if not cls.get_need_blok(required):
                raise BlokManagerException(
                    "Not %s required bloks found" % required)

            cls.bloks[required].required_by.append(blok)

        for optional in cls.bloks[blok].optional:
            if cls.get_need_blok(optional):
                cls.bloks[optional].optional_by.append(blok)

        for conditional in cls.bloks[blok].conditional:
            cls.bloks[conditional].conditional_by.append(blok)

        for conflicting in cls.bloks[blok].conflicting:
            cls.bloks[conflicting].conflicting_by.append(blok)

    @classmethod
    def get_need_blok(cls, blok):
        if cls.has(blok):
            return True

        if blok not in cls.bloks:
            return False

        cls.get_need_blok_linked_bloks(blok)
        cls.ordered_bloks.append(blok)
        EnvironmentManager.set('current_blok', blok)

        if not ImportManager.has(blok):
            # Import only if not exist don't reload here
            mod = ImportManager.add(blok)
            mod.imports()
        else:
            mod = ImportManager.get(blok)
            mod.reload()

        if cls.bloks[blok].autoinstall:
            cls.auto_install.append(blok)

        return True

    @classmethod
    @log(logger)
    def load(cls, entry_points=('bloks',)):
        """ Load all the bloks and import them

        :param entry_points: Use by ``iter_entry_points`` to get the blok
        :exception: BlokManagerException
        """
        if not entry_points:
            raise BlokManagerException("The entry_points mustn't be empty")

        cls.entry_points = entry_points

        if EnvironmentManager.get('current_blok'):
            while EnvironmentManager.get('current_blok'):
                sleep(0.1)

        EnvironmentManager.set('current_blok', 'start')

        bloks = []
        for entry_point in entry_points:
            count = 0
            for i in iter_entry_points(entry_point):
                count += 1
                blok = i.load()
                blok.required_by = []
                blok.optional_by = []
                blok.conditional_by = []
                blok.conflicting_by = []
                cls.set(i.name, blok)
                blok.name = i.name
                bloks.append((blok.priority, i.name))

            if not count:
                raise BlokManagerException(
                    "Invalid bloks group %r" % entry_point)

        # Empty the ordered blok to reload it depending on the priority
        cls.ordered_bloks = []
        bloks.sort()

        try:
            while bloks:
                blok = bloks.pop(0)[1]
                cls.get_need_blok(blok)

        finally:
            EnvironmentManager.set('current_blok', None)

    @classmethod
    def getPath(cls, blok):
        """ Return the path of the blok

        :param blok: blok name in ``ordered_bloks``
        :rtype: absolute path
        """
        blok = cls.get(blok)
        return dirname(modules[blok.__module__].__file__)

    @classmethod
    def add_importer(cls, key, cls_name):
        """ Add a new importer

        :param key: key of the importer
        :param cls_name: name of the model to import
        """
        cls.importers[key] = cls_name

    @classmethod
    def has_importer(cls, key):
        """ Check if an importer  """
        return True if key in cls.importers else False

    @classmethod
    def get_importer(cls, key):
        """ Get the importer class name

        :param key: key of the importer
        :rtype: name of the model to import
        :exception: BlokManagerException
        """
        if not cls.has_importer(key):
            raise BlokManagerException(
                "No importer found for the key %r" % key)

        return cls.importers[key]


class Blok:
    """ Super class for all the bloks

    define the default value for:

    * priority: order to load the blok
    * required: list of the bloks needed to install this blok
    * optional: list of the bloks to be installed if present in the blok list
    * conditional: if all the bloks of this list are installed then install
      this blok
    """

    autoinstall = False
    priority = 100
    required = []
    optional = []
    conditional = []
    conflicting = []
    name = None  # filled by the BlokManager
    author = ''
    logo = ''

    def __init__(self, registry):
        self.registry = registry

    @classmethod
    def import_declaration_module(cls):
        """ Do the python import for the Declaration of the model or other
        """

    def update(self, latest_version):
        """ Call at the installation or update

        :param latest_version: latest version installed, if the blok have not
                               been installing the latest_version will be None
        """

    def pre_migration(self, latest_version):
        """Call at update, before the automigration

        .. warning::

            You can not use the ORM

        :param latest_version: latest version installed, if the blok have not
                               been installing the latest_version will be None
        """

    def post_migration(self, latest_version):
        """Call at update, after the automigration

        :param latest_version: latest version installed, if the blok have not
                               been installing the latest_version will be None
        """

    def uninstall(self):
        """ Call at the uninstallation
        """

    def load(self):
        """ Call at the launch of the application
        """

    def import_file(self, importer_name, model, *file_path, **kwargs):
        """ Import data file

        :param importer_name: Name of the importer (need installation of the
                              Blok which have the importer)
        :param model: Model of the data to import
        :param \*file_path: relative path of the path in this Blok
        :param \*\*kwargs: Option for the importer
        :rtype: return dict of result
        """
        blok_path = BlokManager.getPath(self.name)
        _file = join(blok_path, *file_path)
        logger.info("import %r file: %r", importer_name, _file)
        Importer = self.registry.get(BlokManager.get_importer(importer_name))
        file_to_import = None
        with open(_file, 'rb') as fp:
            file_to_import = fp.read()

        importer = Importer.insert(
            model=model, file_to_import=file_to_import, **kwargs)
        res = importer.run(self.name)
        logger.info("Create %d entries, Update %d entries",
                    len(res['created_entries']), len(res['updated_entries']))
        if res['error_found']:
            for error in res['error_found']:
                logger.error(error)
        else:
            importer.delete()

        return res



# This file is a part of the AnyBlok project
#
#    Copyright (C) 2014 Jean-Sebastien SUZANNE <jssuzanne@anybox.fr>
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file,You can
# obtain one at http://mozilla.org/MPL/2.0/.
from anyblok import Declarations
from anyblok.column import String, Json, Boolean
from ..exceptions import ParameterException


register = Declarations.register
System = Declarations.Model.System


@register(Declarations.Model.System)
class Parameter:
    """System Parameter"""

    key = String(primary_key=True)
    value = Json(nullable=False)
    multi = Boolean(default=False)

    @classmethod
    def set(cls, key, value):
        """ Insert or Update parameter for the key

        :param key: key to save
        :param value: value to save
        """
        multi = False
        if not isinstance(value, dict):
            value = {'value': value}
        else:
            multi = True

        if cls.is_exist(key):
            param = cls.from_primary_keys(key=key)
            param.update(value=value, multi=multi)
        else:
            cls.insert(key=key, value=value, multi=multi)

    @classmethod
    def is_exist(cls, key):
        """ Check if one parameter exist for the key

        :param key: key to check
        :rtype: Boolean, True if exist
        """
        query = cls.query().filter(cls.key == key)
        return True if query.count() else False

    @classmethod
    def get(cls, key):
        """ Return the value of the key

        :param key: key to check
        :rtype: return value
        :exception: ExceptionParameter
        """
        if not cls.is_exist(key):
            raise ParameterException(
                "unexisting key %r" % key)

        param = cls.from_primary_keys(key=key)
        if param.multi:
            return param.value

        return param.value['value']

    @classmethod
    def pop(cls, key):
        """Remove return the value of the key

        :param key: key to check
        :rtype: return value
        :exception: ExceptionParameter
        """
        if not cls.is_exist(key):
            raise ParameterException(
                "unexisting key %r" % key)

        param = cls.from_primary_keys(key=key)
        if param.multi:
            res = param.value
        else:
            res = param.value['value']

        param.delete()
        return res



# This file is a part of the AnyBlok project
#
#    Copyright (C) 2015 Georges Racinet <gracinet@anybox.fr>
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file,You can
# obtain one at http://mozilla.org/MPL/2.0/.
"""Query objects that are rewrapped with permission filtering."""


class QueryWithNoResults:
    """Gives an empty result set without even calling the database

    Useful in cases where the authz system knows beforehand that there won't
    be any result. This can happen for instance if the policy is about the
    model, instead of individual records.
    """
    def count(self):
        return 0

    def all(self):
        return []

    def first(self):
        return None  # TODO exc ?


QUERY_WITH_NO_RESULTS = QueryWithNoResults()


class PostFilteredQuery:
    """Query class returned by the authorization system.

    It takes into account various postfiltering scenarios. Even for cases
    where there is no postfiltering, it's worthwile encapsulate, so that
    downstream code does not rely on methods that would fail if a postfiltering
    policy were to be used.
    """

    def __init__(self, query, postfilters):
        self.query = query
        self.postfilters = postfilters

    def count(self):
        if self.postfilters:
            # TODO add policy information (needs to change __init__)
            raise RuntimeError(
                "Cannot apply count to a permission postfiltered query")
        return self.query.count()

    def filter_one(self, result):
        pfs = self.postfilters
        for rec in result:
            pf = pfs.get(result.__class__)
            if pf is None:
                continue
            if not pf(rec):
                return False

        return True

    def all(self):
        if not self.postfilters:
            return self.query.all()
        return filter(self.filter_one, self.query.all())

    def first(self):
        if not self.postfilters:
            return self.query.first()
        else:
            raise NotImplementedError



# This file is a part of the AnyBlok project
#
#    Copyright (C) 2014 Jean-Sebastien SUZANNE <jssuzanne@anybox.fr>
#    Copyright (C) 2015 Simon ANDRÉ <sandre@anybox.fr>
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file,You can
# obtain one at http://mozilla.org/MPL/2.0/.
from anyblok.tests.testcase import DBTestCase
from anyblok import Declarations
from anyblok.column import Integer, String, Date
from anyblok.relationship import Many2One
from datetime import date


register = Declarations.register
Model = Declarations.Model
Mixin = Declarations.Mixin
Core = Declarations.Core


def single_table_poly():
    @register(Model)
    class Employee:
        id = Integer(primary_key=True)
        type_entity = String(label="Entity type")
        name = String()
        engineer_name = String()
        manager_name = String()

        @classmethod
        def define_mapper_args(cls):
            mapper_args = super(Employee, cls).define_mapper_args()
            mapper_args.update({
                'polymorphic_identity': 'employee',
                'polymorphic_on': cls.type_entity,
            })
            return mapper_args

    @register(Model, tablename=Model.Employee)
    class Engineer(Model.Employee):
        @classmethod
        def define_mapper_args(cls):
            mapper_args = super(Engineer, cls).define_mapper_args()
            mapper_args.update({
                'polymorphic_identity': 'engineer',
            })
            return mapper_args

    @register(Model, tablename=Model.Employee)
    class Manager(Model.Employee):
        @classmethod
        def define_mapper_args(cls):
            mapper_args = super(Manager, cls).define_mapper_args()
            mapper_args.update({
                'polymorphic_identity': 'manager',
            })
            return mapper_args


def multi_table_poly():
    @register(Model)
    class Employee:
        id = Integer(primary_key=True)
        name = String()
        type_entity = String()

        @classmethod
        def define_mapper_args(cls):
            mapper_args = super(Employee, cls).define_mapper_args()
            mapper_args.update({
                'polymorphic_identity': 'employee',
                'polymorphic_on': cls.type_entity,
            })
            return mapper_args

    @register(Model)
    class Engineer(Model.Employee):
        id = Integer(primary_key=True, foreign_key=Model.Employee.use('id'))
        engineer_name = String()

        @classmethod
        def define_mapper_args(cls):
            mapper_args = super(Engineer, cls).define_mapper_args()
            mapper_args.update({
                'polymorphic_identity': 'engineer',
            })
            return mapper_args

    @register(Model)
    class Manager(Model.Employee):
        id = Integer(primary_key=True, foreign_key=Model.Employee.use('id'))
        manager_name = String()

        @classmethod
        def define_mapper_args(cls):
            mapper_args = super(Manager, cls).define_mapper_args()
            mapper_args.update({
                'polymorphic_identity': 'manager',
            })
            return mapper_args


def multi_table_poly_mixins():
    @register(Mixin)
    class MyMixins:
        birthday = Date()

    @register(Model)
    class Employee:
        id = Integer(primary_key=True)
        name = String()
        type_entity = String()

        @classmethod
        def define_mapper_args(cls):
            mapper_args = super(Employee, cls).define_mapper_args()
            mapper_args.update({
                'polymorphic_identity': 'employee',
                'polymorphic_on': cls.type_entity,
            })
            return mapper_args

    @register(Model)
    class Engineer(Model.Employee):
        id = Integer(primary_key=True, foreign_key=Model.Employee.use('id'))
        engineer_name = String()

        @classmethod
        def define_mapper_args(cls):
            mapper_args = super(Engineer, cls).define_mapper_args()
            mapper_args.update({
                'polymorphic_identity': 'engineer',
            })
            return mapper_args

    @register(Model)
    class Manager(Model.Employee, MyMixins):
        id = Integer(primary_key=True, foreign_key=Model.Employee.use('id'))
        manager_name = String()

        @classmethod
        def define_mapper_args(cls):
            mapper_args = super(Manager, cls).define_mapper_args()
            mapper_args.update({
                'polymorphic_identity': 'manager',
            })
            return mapper_args


def multi_table_foreign_key():
    @register(Model)
    class Room:
        id = Integer(primary_key=True)
        name = String()

    @register(Model)
    class Employee:
        id = Integer(primary_key=True)
        name = String()
        type_entity = String()
        room = Many2One(model=Model.Room, one2many='employees')

        @classmethod
        def define_mapper_args(cls):
            mapper_args = super(Employee, cls).define_mapper_args()
            mapper_args.update({
                'polymorphic_identity': 'employee',
                'polymorphic_on': cls.type_entity,
            })
            return mapper_args

    @register(Model)
    class Engineer(Model.Employee):
        id = Integer(primary_key=True, foreign_key=Model.Employee.use('id'))
        engineer_name = String()

        @classmethod
        def define_mapper_args(cls):
            mapper_args = super(Engineer, cls).define_mapper_args()
            mapper_args.update({
                'polymorphic_identity': 'engineer',
            })
            return mapper_args


def multi_table_foreign_key_with_one_define_mapper_args():
    @register(Model)
    class Room:
        id = Integer(primary_key=True)
        name = String()

    @register(Model)
    class Employee:
        id = Integer(primary_key=True)
        name = String()
        type_entity = String()
        room = Many2One(model=Model.Room, one2many='employees')

        @classmethod
        def define_mapper_args(cls):
            mapper_args = super(Employee, cls).define_mapper_args()
            if Model.Employee.__registry_name__ == cls.__registry_name__:
                mapper_args.update({
                    'polymorphic_on': cls.type_entity,
                })

            mapper_args.update({
                'polymorphic_identity': cls.__registry_name__,
            })
            return mapper_args

    @register(Model)
    class Engineer(Model.Employee):
        id = Integer(primary_key=True, foreign_key=Model.Employee.use('id'))
        engineer_name = String()


class TestPolymorphic(DBTestCase):

    def check_registry(self, Model, **kwargs):
        t = Model.insert(name="test", **kwargs)
        t2 = Model.query().filter(
            Model.type_entity == Model.__mapper__.polymorphic_identity).first()
        self.assertEqual(t2, t)

    def test_single_table_poly(self):
        registry = self.init_registry(single_table_poly)
        self.check_registry(registry.Employee)
        self.check_registry(registry.Engineer,
                            engineer_name='An engineer')
        self.check_registry(registry.Manager, manager_name='An manager')

    def test_multi_table_poly(self):
        registry = self.init_registry(multi_table_poly)
        self.check_registry(registry.Employee)
        self.check_registry(registry.Engineer,
                            engineer_name='An engineer')
        self.check_registry(registry.Manager, manager_name='An manager')

    def test_multi_table_poly_mixins(self):
        registry = self.init_registry(multi_table_poly_mixins)
        self.check_registry(registry.Employee)
        self.check_registry(registry.Engineer,
                            engineer_name='An engineer')
        self.check_registry(registry.Manager, manager_name='An manager',
                            birthday=date.today())

    def test_multi_table_foreign_key(self):
        registry = self.init_registry(multi_table_foreign_key)
        room = registry.Room.insert()
        self.check_registry(registry.Employee, room=room)
        self.check_registry(registry.Engineer,
                            engineer_name='An engineer', room=room)

    def test_multi_table_foreign_key2_with_one_define_mapper_args(self):
        registry = self.init_registry(
            multi_table_foreign_key_with_one_define_mapper_args)
        room = registry.Room.insert()
        self.check_registry(registry.Employee, room=room)
        self.check_registry(registry.Engineer,
                            engineer_name='An engineer', room=room)

    def test_query_with_polymorphic(self):
        registry = self.init_registry(multi_table_poly)
        registry.Employee.insert(name='employee')
        registry.Engineer.insert(name='engineer', engineer_name='john')
        registry.Manager.insert(name='manager', manager_name='doe')
        self.assertEqual(registry.Employee.query().count(), 3)
        for mapper in (registry.Engineer,
                       [registry.Engineer, registry.Manager], '*'):
            query = registry.Employee.query().with_polymorphic(mapper)
            query = query.filter(
                registry.Engineer.engineer_name == 'john')
            employee = query.one()
            self.assertTrue(isinstance(employee, registry.Engineer))

    def test_getFieldType(self):
        registry = self.init_registry(multi_table_poly)
        self.assertEqual(registry.Employee.getFieldType('id'), 'Integer')
        self.assertEqual(registry.Employee.getFieldType('name'), 'String')
        self.assertEqual(registry.Employee.getFieldType('type_entity'),
                         'String')
        self.assertEqual(registry.Engineer.getFieldType('id'), 'Integer')
        self.assertEqual(registry.Engineer.getFieldType('name'), 'String')
        self.assertEqual(registry.Engineer.getFieldType('type_entity'),
                         'String')
        self.assertEqual(registry.Engineer.getFieldType('engineer_name'),
                         'String')
        self.assertEqual(registry.Manager.getFieldType('id'), 'Integer')
        self.assertEqual(registry.Manager.getFieldType('name'), 'String')
        self.assertEqual(registry.Manager.getFieldType('type_entity'),
                         'String')
        self.assertEqual(registry.Manager.getFieldType('manager_name'),
                         'String')

    def test_getFieldType_with_mixin(self):
        registry = self.init_registry(multi_table_poly_mixins)
        self.assertEqual(registry.Employee.getFieldType('id'), 'Integer')
        self.assertEqual(registry.Employee.getFieldType('name'), 'String')
        self.assertEqual(registry.Employee.getFieldType('type_entity'),
                         'String')
        self.assertEqual(registry.Engineer.getFieldType('id'), 'Integer')
        self.assertEqual(registry.Engineer.getFieldType('name'), 'String')
        self.assertEqual(registry.Engineer.getFieldType('type_entity'),
                         'String')
        self.assertEqual(registry.Engineer.getFieldType('engineer_name'),
                         'String')
        self.assertEqual(registry.Manager.getFieldType('id'), 'Integer')
        self.assertEqual(registry.Manager.getFieldType('name'), 'String')
        self.assertEqual(registry.Manager.getFieldType('type_entity'),
                         'String')
        self.assertEqual(registry.Manager.getFieldType('manager_name'),
                         'String')
        self.assertEqual(registry.Manager.getFieldType('birthday'),
                         'Date')

    def test_getFieldType_with_relationship(self):
        registry = self.init_registry(multi_table_foreign_key)
        self.assertEqual(registry.Employee.getFieldType('id'), 'Integer')
        self.assertEqual(registry.Employee.getFieldType('name'), 'String')
        self.assertEqual(registry.Employee.getFieldType('type_entity'),
                         'String')
        self.assertEqual(registry.Employee.getFieldType('room'), 'Many2One')
        self.assertEqual(registry.Engineer.getFieldType('id'), 'Integer')
        self.assertEqual(registry.Engineer.getFieldType('name'), 'String')
        self.assertEqual(registry.Engineer.getFieldType('type_entity'),
                         'String')
        self.assertEqual(registry.Engineer.getFieldType('engineer_name'),
                         'String')
        self.assertEqual(registry.Engineer.getFieldType('room'), 'Many2One')

    def test_field_description(self):
        registry = self.init_registry(multi_table_poly)
        fd_employee = list(registry.Employee.fields_description().keys())
        fd_employee.sort()
        fd_engineer = list(registry.Engineer.fields_description().keys())
        fd_engineer.sort()
        fd_manager = list(registry.Manager.fields_description().keys())
        fd_manager.sort()
        self.assertEqual(fd_employee, ['id', 'name', 'type_entity'])
        self.assertEqual(
            fd_engineer, ['engineer_name', 'id', 'name', 'type_entity'])
        self.assertEqual(
            fd_manager, ['id', 'manager_name', 'name', 'type_entity'])



# This file is a part of the AnyBlok project
#
#    Copyright (C) 2014 Jean-Sebastien SUZANNE <jssuzanne@anybox.fr>
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file,You can
# obtain one at http://mozilla.org/MPL/2.0/.
from graphviz.dot import Digraph


class BaseSchema:
    """ Common class extended by the type of schema """

    def __init__(self, name, format='png'):
        self.name = name
        self.format = format
        self._nodes = {}
        self._edges = {}
        self.count = 0

    def add_edge(self, cls_1, cls_2, attr=None):
        """ Add new edge between 2 node

        ::

            dot.add_edge(node1, node2)

        :param cls_1: node (string or object) for the from
        :param cls_2: node (string or object) for the to
        :paam attr: attribute of the edge
        """
        cls_1 = cls_1 if isinstance(cls_1, str) else cls_1.name
        cls_2 = cls_2 if isinstance(cls_2, str) else cls_2.name
        self.count += 1
        self._edges["%s_%s_2_%d" % (cls_1, cls_2, self.count)] = {
            'from': cls_1,
            'to': cls_2,
            'attr': {} if attr is None else attr
        }

    def render(self):
        """Call graphviz to do the schema """
        self.dot = Digraph(name=self.name, format=self.format,
                           node_attr={'shape': 'record',
                                      'style': 'filled',
                                      'fillcolor': 'gray95'})
        for _, cls in self._nodes.items():
            cls.render(self.dot)

        for _, edge in self._edges.items():
            self.dot.edge(edge['from'], edge['to'],
                          _attributes=edge['attr'])

    def save(self):
        """ render and create the output file """
        self.render()
        self.dot.render(self.name)


class TableSchema:
    """ Describe one table """

    def __init__(self, name, parent, islabel=False):
        self.name = name
        self.parent = parent
        self.islabel = islabel
        self.column = []

    def render(self, dot):
        """Call graphviz to create the schema """
        if self.islabel:
            label = "{%s}" % self.name
        else:
            column = '\\n'.join(self.column)
            label = "{%s|%s}" % (self.name, column)

        dot.node(self.name, label=label)

    def add_column(self, name, type_, primary_key=False):
        """Add a new column in the table

        :param name: name of the column
        :param type_: type of the column
        :param primary_key: if True, the string PK will be add
        """
        self.column.append("%s%s (%s)" % (
            'PK ' if primary_key else '', name, type_))

    def add_foreign_key(self, node, label=None, nullable=True):
        """ Add a new foreign key

        :param node: node (string or object) of the table linked
        :param label: name of the column of the foreign key
        :param nullable: bool to select the multiplicity of the association
        """
        self.parent.add_foreign_key(self, node, label, nullable)


class SQLSchema(BaseSchema):
    """ Create a schema to display the table model

    ::

        dot = SQLSchema('the name of my schema')
        t1 = dot.add_table('Table 1')
        t1.add_column('c1', 'Integer')
        t1.add_column('c2', 'Integer')
        t2 = dot.add_table('Table 2')
        t2.add_column('c1', 'Integer')
        t2.add_foreign_key(t1, 'c2')
        dot.save()
    """

    def add_table(self, name):
        """ Add a new node TableSchema with column

        :param name: name of the table
        :rtype: return the instance of TableSchema
        """
        tmp = TableSchema(name, self)
        self._nodes[name] = tmp
        return tmp

    def add_label(self, name):
        """ Add a new node TableSchema without column

        :param name: name of the table
        :rtype: return the instance of TableSchema
        """
        tmp = TableSchema(name, self, islabel=True)
        self._nodes[name] = tmp
        return tmp

    def get_table(self, name):
        """ Return the instance of TableSchema linked with the name of table

        :param name: name of the table
        :rtype: return the instance of TableSchema
        """
        return self._nodes.get(name)

    def add_foreign_key(self, cls_1, cls_2, label=None, nullable=False):
        multiplicity = "0..1" if nullable else "1"
        hlabel = '%s (%s)' % (label, multiplicity) if label else multiplicity

        self.add_edge(cls_1, cls_2, attr={
            'arrowhead': "none",
            'headlabel': hlabel,
        })


class ClassSchema:
    """ Use to display a class """

    def __init__(self, name, parent, islabel=False):
        self.name = name
        self.parent = parent
        self.islabel = islabel
        self.properties = []
        self.column = []
        self.method = []

    def extend(self, node):
        """ add an edge with extend shape to the node

        :param node: node (string or object)
        """
        self.parent.add_extend(self, node)

    def strong_agregate(self, node,
                        label_from=None, multiplicity_from=None,
                        label_to=None, multiplicity_to=None):
        """ add an edge with strong agregate shape to the node

        :param node: node (string or object)
        :param label_from: attribute name
        :param multiplicity_from: multiplicity of the attribute
        :param label_to: attribute name
        :param multiplicity_to: multiplicity of the attribute
        """
        self.parent.add_strong_agregation(self, node, label_from,
                                          multiplicity_from, label_to,
                                          multiplicity_to)

    def agregate(self, node,
                 label_from=None, multiplicity_from=None,
                 label_to=None, multiplicity_to=None):
        """ add an edge with agregate shape to the node

        :param node: node (string or object)
        :param label_from: attribute name
        :param multiplicity_from: multiplicity of the attribute
        :param label_to: attribute name
        :param multiplicity_to: multiplicity of the attribute
        """
        self.parent.add_agregation(self, node, label_from, multiplicity_from,
                                   label_to, multiplicity_to)

    def associate(self, node,
                  label_from=None, multiplicity_from=None,
                  label_to=None, multiplicity_to=None):
        """ add an edge with associate shape to the node

        :param node: node (string or object)
        :param label_from: attribute name
        :param multiplicity_from: multiplicity of the attribute
        :param label_to: attribute name
        :param multiplicity_to: multiplicity of the attribute
        """
        self.parent.add_association(self, node, label_from, multiplicity_from,
                                    label_to, multiplicity_to)

    def add_property(self, name):
        """ add a property in the class

        :param name: name of the property
        """
        self.properties.append(name)

    def add_column(self, name):
        """ add a column in the class

        :param name: name of the column
        """
        self.column.append(name)

    def add_method(self, name):
        """ add a method in the class

        :param name: name of the method
        """
        self.method.append(name)

    def render(self, dot):
        """Call graphviz to do the schema """
        if self.islabel:
            label = "{%s}" % self.name
        else:
            properties = '\\n'.join(self.properties)
            column = '\\n'.join(self.column)
            method = '\\n'.join('%s()' % x for x in self.method)
            label = "{%s|%s|%s|%s}" % (self.name, properties, column, method)

        dot.node(self.name, label=label)


class ModelSchema(BaseSchema):
    """ Create a schema to display the UML model

    ::

        dot = ModelSchema('The name of my UML schema')
        cls = dot.add_class('My class')
        cls.add_method('insert')
        cls.add_property('items')
        cls.add_column('my column')
        dot.save()
    """

    def add_class(self, name):
        """ Add a new node ClassSchema with column

        :param name: name of the class
        :rtype: return the instance of ClassSchema
        """
        tmp = ClassSchema(name, self)
        self._nodes[name] = tmp
        return tmp

    def add_label(self, name):
        """ Return the instance of ClassSchema linked with the name of class

        :param name: name of the class
        :rtype: return the instance of ClassSchema
        """
        tmp = ClassSchema(name, self, islabel=True)
        self._nodes[name] = tmp
        return tmp

    def get_class(self, name):
        """ Add a new node ClassSchema without column

        :param name: name of the class
        :rtype: return the instance of ClassSchema
        """
        return self._nodes.get(name)

    def add_extend(self, cls_1, cls_2):
        self.add_edge(cls_1, cls_2, attr={
            'dir': 'back',
            'arrowtail': 'empty',
        })

    def add_agregation(self, cls_1, cls_2,
                       label_from=None, multiplicity_from=None,
                       label_to=None, multiplicity_to=None):
        label_from, label_to = self.format_label(
            label_from, multiplicity_from, label_to, multiplicity_to)
        self.add_edge(cls_1, cls_2, attr={
            'dir': 'back',
            'arrowtail': 'odiamond',
            'headlabel': label_from,
            'taillabel': label_to,
        })

    def add_strong_agregation(self, cls_1, cls_2,
                              label_from=None, multiplicity_from=None,
                              label_to=None, multiplicity_to=None):
        label_from, label_to = self.format_label(
            label_from, multiplicity_from, label_to, multiplicity_to)
        self.add_edge(cls_1, cls_2, attr={
            'dir': 'back',
            'arrowtail': 'diamond',
            'headlabel': label_from,
            'taillabel': label_to,
        })

    def format_label(self, label_from, multiplicity_from, label_to,
                     multiplicity_to):
        def _format_label(label, multiplicity):
            if label:
                if multiplicity:
                    return '%s (%s)' % (label, multiplicity)

                return label
            else:
                if multiplicity:
                    return multiplicity

                return

        return (
            _format_label(label_from, multiplicity_from),
            _format_label(label_to, multiplicity_to),
        )

    def add_association(self, cls_1, cls_2,
                        label_from=None, multiplicity_from=None,
                        label_to=None, multiplicity_to=None):
        label_from, label_to = self.format_label(
            label_from, multiplicity_from, label_to, multiplicity_to)
        self.add_edge(cls_1, cls_2, attr={
            'arrowhead': "none",
            'headlabel': label_from,
            'taillabel': label_to,
        })



# This file is a part of the AnyBlok project
#
#    Copyright (C) 2014 Jean-Sebastien SUZANNE <jssuzanne@anybox.fr>
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file,You can
# obtain one at http://mozilla.org/MPL/2.0/.
from anyblok.tests.testcase import TestCase
from anyblok.registry import RegistryManager
from anyblok import Declarations
from anyblok.environment import EnvironmentManager
register = Declarations.register
unregister = Declarations.unregister
Core = Declarations.Core


class OneInterface:
    pass


class TestCoreInterfaceCoreBase(TestCase):

    _corename = 'Base'

    @classmethod
    def setUpClass(cls):
        super(TestCoreInterfaceCoreBase, cls).setUpClass()
        RegistryManager.init_blok('testCore' + cls._corename)
        EnvironmentManager.set('current_blok', 'testCore' + cls._corename)

    @classmethod
    def tearDownClass(cls):
        super(TestCoreInterfaceCoreBase, cls).tearDownClass()
        EnvironmentManager.set('current_blok', None)
        del RegistryManager.loaded_bloks['testCore' + cls._corename]

    def setUp(self):
        super(TestCoreInterfaceCoreBase, self).setUp()
        blokname = 'testCore' + self._corename
        RegistryManager.loaded_bloks[blokname]['Core'][self._corename] = []

    def assertInCore(self, *args):
        blokname = 'testCore' + self._corename
        blok = RegistryManager.loaded_bloks[blokname]
        self.assertEqual(len(blok['Core'][self._corename]), len(args))
        for cls_ in args:
            hasCls = cls_ in blok['Core'][self._corename]
            self.assertEqual(hasCls, True)

    def assertInRemoved(self, cls):
        core = RegistryManager.loaded_bloks['testCoreBase']['removed']
        if cls in core:
            return True

        self.fail('Not in removed')

    def test_add_interface(self):
        register(Core, cls_=OneInterface, name_='Base')
        self.assertEqual('Core', Core.Base.__declaration_type__)
        self.assertInCore(OneInterface)
        dir(Declarations.Core.Base)

    def test_add_interface_with_decorator(self):

        @register(Core)
        class Base:
            pass

        self.assertEqual('Core', Core.Base.__declaration_type__)
        self.assertInCore(Base)

    def test_add_two_interface(self):

        register(Core, cls_=OneInterface, name_="Base")

        @register(Core)
        class Base:
            pass

        self.assertInCore(OneInterface, Base)

    def test_remove_interface_with_1_cls_in_registry(self):

        register(Core, cls_=OneInterface, name_="Base")
        self.assertInCore(OneInterface)
        unregister(Core.Base, OneInterface)
        self.assertInCore(OneInterface)
        self.assertInRemoved(OneInterface)

    def test_remove_interface_with_2_cls_in_registry(self):

        register(Core, cls_=OneInterface, name_="Base")

        @register(Core)
        class Base:
            pass

        self.assertInCore(OneInterface, Base)
        unregister(Core.Base, OneInterface)
        self.assertInCore(Base, OneInterface)
        self.assertInRemoved(OneInterface)



# -*- coding: utf-8 -*-
# This file is a part of the AnyBlok project
#
#    Copyright (C) 2014 Jean-Sebastien SUZANNE <jssuzanne@anybox.fr>
#    Copyright (C) 2015 Pierre Verkest <pverkest@anybox.fr>
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file,You can
# obtain one at http://mozilla.org/MPL/2.0/.
from anyblok.tests.testcase import DBTestCase
from anyblok import Declarations
from anyblok.mapper import ModelAttributeException
from anyblok.column import Integer, String
from anyblok.relationship import Many2Many

register = Declarations.register
Model = Declarations.Model
Mixin = Declarations.Mixin


def _complete_many2many(**kwargs):

    @register(Model)
    class Address:

        id = Integer(primary_key=True)
        street = String()
        zip = String()
        city = String()

    @register(Model)
    class Person:

        name = String(primary_key=True)
        addresses = Many2Many(model=Model.Address,
                              join_table="join_addresses_by_persons",
                              remote_columns="id", local_columns="name",
                              m2m_remote_columns='a_id',
                              m2m_local_columns='p_name',
                              many2many="persons")


def _minimum_many2many(**kwargs):

    @register(Model)
    class Address:

        id = Integer(primary_key=True)
        street = String()
        zip = String()
        city = String()

    @register(Model)
    class Person:

        name = String(primary_key=True)
        addresses = Many2Many(model=Model.Address)


def _minimum_many2many_by_mixin(**kwargs):

    @register(Model)
    class Address:

        id = Integer(primary_key=True)
        street = String()
        zip = String()
        city = String()

    @register(Mixin)
    class MixinM2M:
        addresses = Many2Many(model=Model.Address)

    @register(Model)
    class Person(Mixin.MixinM2M):

        name = String(primary_key=True)


def _many2many_with_str_model(**kwargs):

    @register(Model)
    class Address:

        id = Integer(primary_key=True)
        street = String()
        zip = String()
        city = String()

    @register(Model)
    class Person:

        name = String(primary_key=True)
        addresses = Many2Many(model='Model.Address')


def auto_detect_two_primary_keys(**kwargs):

    @register(Model)
    class Address:

        id = Integer(primary_key=True)
        id2 = Integer(primary_key=True)

    @register(Model)
    class Person:

        name = String(primary_key=True)
        addresses = Many2Many(model=Model.Address)


def unexisting_remote_columns(**kwargs):

    @register(Model)
    class Address:

        id = Integer(primary_key=True)

    @register(Model)
    class Person:

        name = String(primary_key=True)
        addresses = Many2Many(model=Model.Address, remote_columns="id2")


def reuse_many2many_table(**kwargs):

    @register(Model)
    class Address:

        id = Integer(primary_key=True)

    @register(Model)
    class Person:

        name = String(primary_key=True)
        addresses = Many2Many(model=Model.Address)

    @register(Model)  # noqa
    class Address:

        id = Integer(primary_key=True)
        persons = Many2Many(model=Model.Person,
                            join_table='join_person_and_address')


class TestMany2Many(DBTestCase):

    def test_complete_many2many(self):
        registry = self.init_registry(_complete_many2many)

        address_exist = hasattr(registry.Person, 'addresses')
        self.assertTrue(address_exist)

        m2m_tables_exist = hasattr(registry, 'many2many_tables')
        self.assertTrue(m2m_tables_exist)

        jt = registry.declarativebase.metadata.tables
        join_table_exist = 'join_addresses_by_persons' in jt
        self.assertEqual(join_table_exist, True)

        self.assertEqual(
            len(jt['join_addresses_by_persons'].primary_key.columns),
            2)
        self.assertTrue('a_id' in jt[
            'join_addresses_by_persons'].primary_key.columns)
        self.assertTrue('p_name' in jt[
            'join_addresses_by_persons'].primary_key.columns)

        address = registry.Address.insert(
            street='14-16 rue soleillet', zip='75020', city='Paris')

        person = registry.Person.insert(name="Jean-sébastien SUZANNE")

        person.addresses.append(address)

        self.assertEqual(address.persons, [person])

    def test_complete_many2many_expire_field(self):
        registry = self.init_registry(_complete_many2many)
        self.assertIn(
            ('x2m', 'addresses', 'persons'),
            registry.expire_attributes['Model.Person']['addresses'])
        self.assertIn(
            ('x2m', 'persons', 'addresses'),
            registry.expire_attributes['Model.Address']['persons'])

    def test_minimum_many2many(self):
        registry = self.init_registry(_minimum_many2many)

        address_exist = hasattr(registry.Person, 'addresses')
        self.assertTrue(address_exist)

        m2m_tables_exist = hasattr(registry, 'many2many_tables')
        self.assertTrue(m2m_tables_exist)

        jt = registry.declarativebase.metadata.tables
        join_table_exist = 'join_person_and_address' in jt
        self.assertEqual(join_table_exist, True)

        address = registry.Address.insert(
            street='14-16 rue soleillet', zip='75020', city='Paris')

        person = registry.Person.insert(name="Jean-sébastien SUZANNE")

        person.addresses.append(address)

        self.assertEqual(person.addresses, [address])

    def test_many2many_with_str_model(self):
        registry = self.init_registry(_many2many_with_str_model)

        address_exist = hasattr(registry.Person, 'addresses')
        self.assertTrue(address_exist)

        m2m_tables_exist = hasattr(registry, 'many2many_tables')
        self.assertTrue(m2m_tables_exist)

        jt = registry.declarativebase.metadata.tables
        join_table_exist = 'join_person_and_address' in jt
        self.assertEqual(join_table_exist, True)

        address = registry.Address.insert(
            street='14-16 rue soleillet', zip='75020', city='Paris')

        person = registry.Person.insert(name="Jean-sébastien SUZANNE")

        person.addresses.append(address)

        self.assertEqual(person.addresses, [address])

    def test_unexisting_remote_columns(self):
        with self.assertRaises(ModelAttributeException):
            self.init_registry(unexisting_remote_columns)

    def test_reuse_many2many_table(self):
        registry = self.init_registry(reuse_many2many_table)

        address = registry.Address.insert()
        person = registry.Person.insert(name="Jean-sébastien SUZANNE")

        person.addresses.append(address)

        self.assertEqual(address.persons, [person])

    def test_declared_in_mixin(self):
        registry = self.init_registry(_minimum_many2many_by_mixin)

        address_exist = hasattr(registry.Person, 'addresses')
        self.assertTrue(address_exist)

        m2m_tables_exist = hasattr(registry, 'many2many_tables')
        self.assertTrue(m2m_tables_exist)

        jt = registry.declarativebase.metadata.tables
        join_table_exist = 'join_person_and_address' in jt
        self.assertEqual(join_table_exist, True)

        address = registry.Address.insert(
            street='14-16 rue soleillet', zip='75020', city='Paris')

        person = registry.Person.insert(name="Jean-sébastien SUZANNE")

        person.addresses.append(address)

        self.assertEqual(person.addresses, [address])

    def test_declared_in_mixin_inherit_by_two_models(self):
        def add_in_registry():
            _minimum_many2many_by_mixin()

            @register(Model)
            class Person2(Mixin.MixinM2M):

                name = String(primary_key=True)

        registry = self.init_registry(add_in_registry)

        address_exist = hasattr(registry.Person, 'addresses')
        self.assertTrue(address_exist)
        address_exist = hasattr(registry.Person2, 'addresses')
        self.assertTrue(address_exist)

        m2m_tables_exist = hasattr(registry, 'many2many_tables')
        self.assertTrue(m2m_tables_exist)

        jt = registry.declarativebase.metadata.tables
        join_table_exist = 'join_person_and_address' in jt
        self.assertEqual(join_table_exist, True)
        join_table_exist = 'join_person2_and_address' in jt
        self.assertEqual(join_table_exist, True)

    def test_comlet_with_multi_primary_keys_remote_and_local(self):

        def add_in_registry():

            @register(Model)
            class Test:
                id = Integer(primary_key=True)
                id2 = String(primary_key=True)

            @register(Model)
            class Test2:
                id = Integer(primary_key=True)
                id2 = String(primary_key=True)
                test = Many2Many(model=Model.Test,
                                 join_table="join_test_and_test2",
                                 remote_columns=['id', 'id2'],
                                 m2m_remote_columns=['t2_id', 't2_id2'],
                                 local_columns=['id', 'id2'],
                                 m2m_local_columns=['t1_id', 't1_id2'],
                                 many2many="test2")

        registry = self.init_registry(add_in_registry)
        t1 = registry.Test.insert(id2="test1")
        t2 = registry.Test2.insert(id2="test2")
        t2.test.append(t1)
        self.assertIs(t1.test2[0], t2)

        jt = registry.declarativebase.metadata.tables
        self.assertEqual(
            len(jt['join_test_and_test2'].primary_key.columns),
            4)
        self.assertTrue('t1_id' in jt[
            'join_test_and_test2'].primary_key.columns)
        self.assertTrue('t1_id2' in jt[
            'join_test_and_test2'].primary_key.columns)
        self.assertTrue('t2_id' in jt[
            'join_test_and_test2'].primary_key.columns)
        self.assertTrue('t2_id2' in jt[
            'join_test_and_test2'].primary_key.columns)

    def test_comlet_with_multi_primary_keys_remote(self):

        def add_in_registry():

            @register(Model)
            class Test:
                id = Integer(primary_key=True)
                id2 = String(primary_key=True)

            @register(Model)
            class Test2:
                id = Integer(primary_key=True)
                test = Many2Many(model=Model.Test,
                                 join_table="join_test_and_test2",
                                 remote_columns=['id', 'id2'],
                                 m2m_remote_columns=['t2_id', 't2_id2'],
                                 local_columns='id',
                                 m2m_local_columns='t1_id',
                                 many2many="test2")

        registry = self.init_registry(add_in_registry)
        t1 = registry.Test.insert(id2="test1")
        t2 = registry.Test2.insert()
        t2.test.append(t1)
        self.assertIs(t1.test2[0], t2)

    def test_comlet_with_multi_primary_keys_local(self):

        def add_in_registry():

            @register(Model)
            class Test:
                id = Integer(primary_key=True)

            @register(Model)
            class Test2:
                id = Integer(primary_key=True)
                id2 = String(primary_key=True)
                test = Many2Many(model=Model.Test,
                                 join_table="join_test_and_test2",
                                 remote_columns='id',
                                 m2m_remote_columns='t2_id',
                                 local_columns=['id', 'id2'],
                                 m2m_local_columns=['t1_id', 't1_id2'],
                                 many2many="test2")

        registry = self.init_registry(add_in_registry)
        t1 = registry.Test.insert()
        t2 = registry.Test2.insert(id2='test2')
        t2.test.append(t1)
        self.assertIs(t1.test2[0], t2)

    def test_minimum_with_multi_primary_keys_remote_and_local(self):

        def add_in_registry():

            @register(Model)
            class Test:
                id = Integer(primary_key=True)
                id2 = String(primary_key=True)

            @register(Model)
            class Test2:
                id = Integer(primary_key=True)
                id2 = String(primary_key=True)
                test = Many2Many(model=Model.Test, many2many="test2")

        registry = self.init_registry(add_in_registry)
        t1 = registry.Test.insert(id2="test1")
        t2 = registry.Test2.insert(id2="test2")
        t2.test.append(t1)
        self.assertIs(t1.test2[0], t2)

    def test_minimum_with_multi_primary_keys_remote(self):

        def add_in_registry():

            @register(Model)
            class Test:
                id = Integer(primary_key=True)
                id2 = String(primary_key=True)

            @register(Model)
            class Test2:
                id = Integer(primary_key=True)
                test = Many2Many(model=Model.Test, many2many="test2")

        registry = self.init_registry(add_in_registry)
        t1 = registry.Test.insert(id2="test1")
        t2 = registry.Test2.insert()
        t2.test.append(t1)
        self.assertIs(t1.test2[0], t2)

    def test_minimum_with_multi_primary_keys_local(self):

        def add_in_registry():

            @register(Model)
            class Test:
                id = Integer(primary_key=True)

            @register(Model)
            class Test2:
                id = Integer(primary_key=True)
                id2 = String(primary_key=True)
                test = Many2Many(model=Model.Test, many2many="test2")

        registry = self.init_registry(add_in_registry)
        t1 = registry.Test.insert()
        t2 = registry.Test2.insert(id2="test2")
        t2.test.append(t1)
        self.assertIs(t1.test2[0], t2)



# -*- coding: utf-8 -*-
# This file is a part of the AnyBlok project
#
#    Copyright (C) 2014 Jean-Sebastien SUZANNE <jssuzanne@anybox.fr>
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file,You can
# obtain one at http://mozilla.org/MPL/2.0/.
from anyblok.tests.testcase import TestCase, DBTestCase
from anyblok.registry import RegistryManager
from anyblok.config import Configuration
from anyblok.blok import BlokManager, Blok
from anyblok.column import Integer
from threading import Thread


class Test:
    pass


class TestTest:
    pass


class TestTestTest:
    pass


class TestRegistry(DBTestCase):

    def tearDown(self):
        super(TestRegistry, self).tearDown()
        for cls in (Test, TestTest):
            if hasattr(cls, 'Test'):
                delattr(cls, 'Test')

    def test_get_model(self):
        registry = self.init_registry(None)
        System = registry.get('Model.System')
        self.assertEqual(registry.System, System)

    def test_get_the_same_registry(self):
        registry = self.init_registry(None)
        registry2 = self.getRegistry()
        self.assertEqual(registry, registry2)

    def test_reload(self):
        registry = self.init_registry(None)
        bloks_before_reload = registry.System.Blok.query('name').filter(
            registry.System.Blok.state == 'installed').all()
        registry.reload()
        bloks_after_reload = registry.System.Blok.query('name').filter(
            registry.System.Blok.state == 'installed').all()
        self.assertEqual(bloks_before_reload, bloks_after_reload)

    def test_get_bloks_to_load(self):
        registry = self.init_registry(None)
        bloks = registry.get_bloks_to_load()
        self.assertIn('anyblok-core', bloks)

    def test_load_entry(self):
        registry = self.init_registry(None)
        registry.loaded_registries['entry_names'] = []
        RegistryManager.loaded_bloks['blok'] = {
            'entry': {
                'registry_names': ['key'],
                'key': {'properties': {'property': True},
                        'bases': [TestCase]},
            },
        }
        registry.load_entry('blok', 'entry')
        self.assertEqual(registry.loaded_registries['key'],
                         {'properties': {'property': True},
                          'bases': [TestCase]})

    def test_load_core(self):
        RegistryManager.loaded_bloks['blok'] = {
            'Core': {'Session': [TestCase]},
        }
        registry = self.init_registry(None)
        registry.load_core('blok', 'Session')
        self.assertIn(TestCase, registry.loaded_cores['Session'])

    def test_load_blok(self):

        class BlokTest(Blok):
            pass

        registry = self.init_registry(None)
        BlokManager.bloks['blok'] = BlokTest
        RegistryManager.loaded_bloks['blok'] = {
            'Core': {'Session': [TestCase],
                     'Base': [],
                     'SqlBase': [],
                     'SqlViewBase': [],
                     'Query': [],
                     'InstrumentedList': []},
            'removed': [],
            'properties': {},
        }
        for entry in RegistryManager.declared_entries:
            RegistryManager.loaded_bloks['blok'][entry] = {
                'registry_names': []}

        registry.load_blok('blok', False, [])
        self.assertIn(TestCase, registry.loaded_cores['Session'])

    def check_added_in_regisry(self, registry):
        self.assertEqual(registry.Test, Test)
        self.assertEqual(registry.Test.Test, TestTest)
        self.assertEqual(registry.Test.Test.Test, TestTestTest)

    def test_add_in_registry_1(self):
        registry = self.init_registry(None)
        registry.add_in_registry('Declarations.Test', Test)
        registry.add_in_registry('Declarations.Test.Test', TestTest)
        registry.add_in_registry('Declarations.Test.Test.Test', TestTestTest)
        self.check_added_in_regisry(registry)

    def test_add_in_registry_2(self):
        registry = self.init_registry(None)
        registry.add_in_registry('Declarations.Test', Test)
        registry.add_in_registry('Declarations.Test.Test.Test', TestTestTest)
        registry.add_in_registry('Declarations.Test.Test', TestTest)
        self.check_added_in_regisry(registry)

    def test_add_in_registry_3(self):
        registry = self.init_registry(None)
        registry.add_in_registry('Declarations.Test.Test', TestTest)
        registry.add_in_registry('Declarations.Test', Test)
        registry.add_in_registry('Declarations.Test.Test.Test', TestTestTest)
        self.check_added_in_regisry(registry)

    def test_add_in_registry_4(self):
        registry = self.init_registry(None)
        registry.add_in_registry('Declarations.Test.Test', TestTest)
        registry.add_in_registry('Declarations.Test.Test.Test', TestTestTest)
        registry.add_in_registry('Declarations.Test', Test)
        self.check_added_in_regisry(registry)

    def test_add_in_registry_5(self):
        registry = self.init_registry(None)
        registry.add_in_registry('Declarations.Test.Test.Test', TestTestTest)
        registry.add_in_registry('Declarations.Test', Test)
        registry.add_in_registry('Declarations.Test.Test', TestTest)
        self.check_added_in_regisry(registry)

    def test_add_in_registry_6(self):
        registry = self.init_registry(None)
        registry.add_in_registry('Declarations.Test.Test.Test', TestTestTest)
        registry.add_in_registry('Declarations.Test.Test', TestTest)
        registry.add_in_registry('Declarations.Test', Test)
        self.check_added_in_regisry(registry)

    def test_registry_db_exist(self):
        self.assertTrue(Configuration.get('Registry').db_exists(
            db_name=Configuration.get('db_name')))

    def test_registry_db_unexist(self):
        self.assertFalse(Configuration.get('Registry').db_exists(
            db_name='wrong_db_name'))


class TestRegistry2(DBTestCase):

    def add_model(self):

        from anyblok import Declarations

        register = Declarations.register
        Model = Declarations.Model

        @register(Model)
        class Test:

            id = Integer(primary_key=True)
            val = Integer(default=0)

            @classmethod
            def _precommit_hook(cls):
                for t in cls.query().all():
                    t.val += t.id

            def add_precommit_hook(self):
                self.precommit_hook('_precommit_hook')

            @classmethod
            def add_cl_precommit_hook(cls):
                cls.precommit_hook('_precommit_hook')

    def test_precommit_hook(self):
        registry = self.init_registry(self.add_model)
        t1 = registry.Test.insert()
        t1.add_precommit_hook()
        t2 = registry.Test.insert()
        t2.add_precommit_hook()
        self.assertEqual(t1.val, 0)
        self.assertEqual(t2.val, 0)
        registry.commit()
        self.assertEqual(t1.val, t1.id)
        self.assertEqual(t2.val, t2.id)
        registry.commit()
        self.assertEqual(t1.val, t1.id)
        self.assertEqual(t2.val, t2.id)
        registry.Test.add_cl_precommit_hook()
        self.assertEqual(t1.val, t1.id)
        self.assertEqual(t2.val, t2.id)
        registry.commit()
        self.assertEqual(t1.val, 2 * t1.id)
        self.assertEqual(t2.val, 2 * t2.id)
        registry.commit()
        self.assertEqual(t1.val, 2 * t1.id)
        self.assertEqual(t2.val, 2 * t2.id)
        t1.add_precommit_hook()
        self.assertEqual(t1.val, 2 * t1.id)
        self.assertEqual(t2.val, 2 * t2.id)
        registry.commit()
        self.assertEqual(t1.val, 3 * t1.id)
        self.assertEqual(t2.val, 3 * t2.id)

    def test_precommit_hook_in_thread(self):
        registry = self.init_registry(self.add_model)
        t1 = registry.Test.insert()
        t1.add_precommit_hook()
        t2 = registry.Test.insert()
        t2.add_precommit_hook()
        self.assertEqual(t1.val, 0)
        self.assertEqual(t2.val, 0)

        def target():
            registry.commit()

        t = Thread(target=target)
        t.start()
        t.join()

        self.assertEqual(t1.val, 0)
        self.assertEqual(t2.val, 0)
        registry.commit()
        self.assertEqual(t1.val, t1.id)
        self.assertEqual(t2.val, t2.id)

    def define_cls(self, typename='Model', name='Test', val=1, usesuper=False,
                   inherit=None):

        from anyblok import Declarations

        register = Declarations.register
        Type = getattr(Declarations, typename)
        if inherit is None:
            inherit = object
        else:
            inherit = getattr(Declarations.Mixin, inherit)

        @register(Type, name_=name)
        class Test(inherit):

            @classmethod
            def foo(cls):
                if usesuper:
                    return val * super(Test, cls).foo()

                return val

        return Test

    def test_check_define_cls(self):

        def add_in_registry():
            self.define_cls()

        registry = self.init_registry(add_in_registry)
        self.assertEqual(registry.Test.foo(), 1)

    def test_check_define_cls_with_inherit(self):

        def add_in_registry():
            self.define_cls()
            self.define_cls(val=2, usesuper=True)

        registry = self.init_registry(add_in_registry)
        self.assertEqual(registry.Test.foo(), 2)

    def test_check_define_cls_with_inherit_core(self):

        def add_in_registry():
            self.define_cls(typename='Core', name='Base', val=2)
            self.define_cls(val=2, usesuper=True)

        registry = self.init_registry(add_in_registry)
        self.assertEqual(registry.Test.foo(), 4)

    def test_check_define_cls_with_inherit_mixin(self):

        def add_in_registry():
            self.define_cls(typename='Mixin', name='MTest', val=3)
            self.define_cls(val=3, usesuper=True, inherit='MTest')

        registry = self.init_registry(add_in_registry)
        self.assertEqual(registry.Test.foo(), 9)

    def test_check_define_cls_with_inherit2(self):

        def add_in_registry():
            self.define_cls()
            self.define_cls(val=2, usesuper=True)
            self.define_cls(val=2, usesuper=True)

        registry = self.init_registry(add_in_registry)
        self.assertEqual(registry.Test.foo(), 4)

    def test_check_define_cls_with_inherit_core2(self):

        def add_in_registry():
            self.define_cls(typename='Core', name='Base', val=2)
            self.define_cls(typename='Core', name='Base', val=2, usesuper=True)
            self.define_cls(val=2, usesuper=True)

        registry = self.init_registry(add_in_registry)
        self.assertEqual(registry.Test.foo(), 8)

    def test_check_define_cls_with_inherit_mixin2(self):

        def add_in_registry():
            self.define_cls(typename='Mixin', name='MTest', val=3)
            self.define_cls(typename='Mixin', name='MTest', val=3,
                            usesuper=True)
            self.define_cls(val=3, usesuper=True, inherit='MTest')

        registry = self.init_registry(add_in_registry)
        self.assertEqual(registry.Test.foo(), 27)

    def test_remove(self):

        def add_in_registry():
            self.define_cls()
            cls_ = self.define_cls(val=2, usesuper=True)
            self.define_cls(val=2, usesuper=True)
            from anyblok import Declarations
            Declarations.unregister(Declarations.Model.Test, cls_)

        registry = self.init_registry(add_in_registry)
        self.assertEqual(registry.Test.foo(), 2)

    def test_remove_core(self):

        def add_in_registry():
            self.define_cls(typename='Core', name='Base', val=2)
            cls_ = self.define_cls(typename='Core', name='Base', val=2,
                                   usesuper=True)
            self.define_cls(val=2, usesuper=True)
            from anyblok import Declarations
            Declarations.unregister(Declarations.Core.Base, cls_)

        registry = self.init_registry(add_in_registry)
        self.assertEqual(registry.Test.foo(), 4)

    def test_remove_mixin(self):

        def add_in_registry():
            self.define_cls(typename='Mixin', name='MTest', val=3)
            cls_ = self.define_cls(typename='Mixin', name='MTest', val=3,
                                   usesuper=True)
            self.define_cls(val=3, usesuper=True, inherit='MTest')
            from anyblok import Declarations
            Declarations.unregister(Declarations.Mixin.MTest, cls_)

        registry = self.init_registry(add_in_registry)
        self.assertEqual(registry.Test.foo(), 9)



# This file is a part of the AnyBlok project
#
#    Copyright (C) 2014 Jean-Sebastien SUZANNE <jssuzanne@anybox.fr>
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file,You can
# obtain one at http://mozilla.org/MPL/2.0/.
from anyblok import Declarations
from anyblok.column import String
from anyblok.authorization.rule.modelaccess import ModelAccessRule
from anyblok.authorization.rule.attraccess import AttributeAccessRule


register = Declarations.register
Model = Declarations.Model


@register(Model)
class Test2:
    """We'll work on Test2, on which test_blok7 doesn't set any authz policy.
    """
    owner = String()


Declarations.AuthorizationBinding(
    Declarations.Model.Test2,
    ModelAccessRule())

Declarations.AuthorizationBinding(
    Declarations.Model.Test2,
    AttributeAccessRule('owner'),
    permission='Write')

Declarations.AuthorizationBinding(
    Declarations.Model.Test2,
    AttributeAccessRule('owner', model_rule=ModelAccessRule()),
    permission='PermWithModelRule')



from anyblok.blok import Blok
from anyblok.release import version


class mockblok(Blok):
    version = version

    @classmethod
    def import_declaration_module(cls):
        from . import mockfile  # noqa
        from . import mockpackage  # noqa

    @classmethod
    def reload_declaration_module(cls, reload):
        from . import mockfile
        reload(mockfile)
        from . import mockpackage
        reload(mockpackage)
        reload(mockpackage.mockfile1)
        reload(mockpackage.mockfile2)
        reload(mockpackage.submockpackage)
        reload(mockpackage.submockpackage.mockfile1)
        reload(mockpackage.submockpackage.mockfile2)
